from matplotlib import pyplot as plt
#% matplotlib inline
import numpy as np
import pandas as pd

def load_all_events(MSD_instance):
    """
    Uses the MemorySearchDynamics Class to extract out all event arrays
    as well as a single subject's (LTP093 by default)

    ------
    INPUTS
    MSD_instance: an instance of MemorySearchDynamic class
    """
    # GET EVENTS
    all_events = MSD_instance.all_events
    events = all_events[all_events['subject'] == MSD_instance.subjs[0]]
    return all_events, events


def get_all_events_d(all_events):
    """
    Returns a dictionary where each key is a subject and each value is
    the event of that subject
    -------
    INPUTS
    all_events: an arr containing all events
    """
    d = {subj: all_events[all_events['subject'] == subj] for i, subj in enumerate(
        np.unique(all_events['subject']))}
    return d


def list_to_df(df_lists):
    """
    Converts a list of DataFrames into one DataFrame,
    averaging across indexes
    -----
    INPUTS
    df_lists: a list of pd.DataFrames
    """
    concat = pd.concat(df_lists)
    concat = concat.groupby(concat.index).mean()
    return concat


def get_good_events(events, sess, trial):
    """
    Takes in events array and returns out one without any intrusions or
    any repeats. Repeats are removed entirely (both) same with intrusions.

    INPUTS
    ------
    events: an events np.recarray
    sess: The session to apply it over
    Trial: The trial to apply it over
    """
    # Get single sess/trial for subj, only rec words no intrusion
    events = events[(events['session'] == sess)
                    & (events['list'] == trial)
                    & (events['type'] == 'REC_WORD')
                    & (events['intrusion'] == 0)]
    d = dict(pd.Series(events['item_name']).value_counts())
    # Isolate bad words
    bad_words = [v for k, v in enumerate(d) if d[v] > 1]
    # Seperate out the bad words
    if len(bad_words) > 0:
        events = np.array([event for event in events
                           if event['item_name'] not in bad_words])  # .view(np.recarray)
    return events


def get_field(events, fields, sess, trial):
    """
    Gets the difference in the field of a single sess/trial of
    inputted events after filtering it using get_good_events

    INPUTS
    -----
    events: an array of beh events
    sess: session number must not be 0 or >23
    trial: number between 1-24
    """
    events = get_good_events(events, sess, trial)
    # Difference between sequential events
    if len(events) > 1:
        return events[fields]
    else:
        # print 'no events in sess {} trial {}'.format(sess,trial)
        return []


def get_field_diff(events, fields, sess, trial):
    """
    Gets the difference in the field of a single sess/trial of
    inputted events after filtering it using get_good_events

    INPUTS
    -----
    events: an array of beh events
    sess: session number must not be 0 or >23
    trial: number between 1-24
    """
    events = get_good_events(events, sess, trial)
    # Difference between sequential events
    if len(events) > 1:
        # Grab field
        diff = np.diff(events[fields])
        k = len(diff)
        if k > 0:
            return diff
    else:
        # print 'no events in sess {} trial {}'.format(sess,trial)
        return []


def get_field_diff_per_sess(events, sess, fieldname):
    """
    Gets all IRT per session of inputted events
    INPUTS
    -----
    events: an array of beh events
    sess: a session
    """
    # Get all lists from one sess
    session = [get_field_diff(events, fieldname, sess, i) for i in range(1, 25)]
    # Make the index be the len and then we'll group by that
    indx = [len(x) for x in session]
    try:
        df = pd.DataFrame(session, index=indx).groupby(indx).mean()
        return df
    except:
        pass
        # print indx


def get_field_per_subj(events, fieldname):
    """
    Gets ALL FIELDS fieldname of a single subject
    --------
    INPUTS
    events: events for a subject
    """
    all_irts = [get_field_diff_per_sess(events, i, fieldname
                                        ) for i in range(1, 24)]
    return list_to_df(all_irts)


def diff_in_field(events, fields, sess, trial, sp=None):
    """
    Gets the difference in the field of a single sess/trial of
    inputted events after filtering it using get_good_events,
    THIS VERSION WILL BE USED TO INTERACT WITH SERIALPOS SUCH THAT
    WE CAN CONTROL IF WE DO 0, 1, 2, ALL, ETC.

    INPUTS
    -----
    events: an array of beh events
    sess: session number must not be 0 or >23
    trial: number between 1-24
    """
    events = get_good_events(events, sess, trial)
    # Difference between sequential events
    if len(events) > 1:
        # Grab field
        diff = np.diff(events[fields])
        if type(sp) == int:
            try:
                diff = diff[sp]
                return diff
            except:
                return []
        elif type(sp) == str:
            if len(diff) > 0:
                return diff
    else:
        # print 'no events in sess {} trial {}'.format(sess,trial)
        return []


def plot_irt(irt_per_sess, title):
    """
    Plots an IRT curve, one line per len(recall)
    ------
    INPUTS
    irt_per_sess: output of get_irt_per_sess()
    """
    for i in range(len(irt_per_sess)):
        times = irt_per_sess.iloc[i].values
        pos = irt_per_sess.iloc[i].index
        # Remove all those pesky nans
        x = times[~np.isnan(times)]
        pos = pos[:(len(x))] + 1
        plt.plot(pos, x, label='len {}'.format(len(pos)), marker='*')
        plt.legend(bbox_to_anchor=(1.05, 1), loc=2, borderaxespad=0., ncol=2)
        plt.title('Inter-Response Interval ' + title)
        plt.ylabel('Time (ms)')
        plt.xlabel('Recall Serial Position')
    plt.show()


def get_irt_all(all_events_d):
    """
    Gets IRT for all subjects, generates a plot
    ------
    INPUTS
    all_events_d: a dictionary where keys=subj_id, and values=events_subj
    """
    my_list = [get_field_per_subj(all_events_d[subj], 'rectime'
                                  ) for subj in all_events_d.keys()]
    try:
        df = list_to_df(my_list)
        plot_irt(df)
        plt.show()
        return df
    except:
        try:
            return list_to_df(my_list)
        except:
            return my_list


"""butterworth.py script used for consctructing various Butterworth Filters"""
import numpy as np
from scipy.signal import (filtfilt, firwin, sosfiltfilt, butter, tf2zpk)
try:
    from collections import Sequence
except ImportError as e:
    print('Import error from collections import Sequence')
except Exception as e:
    print(e)
from copy import deepcopy
import warnings
from xarray import DataArray
from ptsa.data.timeseries import TimeSeries
from ptsa.data.common import get_axis_index
from ptsa.data.filters import BaseFilter
import traits.api
from Clumsy.signal.timeseriesLF import TimeSeriesLF

__all__ = ['butterworth_filter', 'ButterworthFilter', 'check_stability', 'FIR_bandpass_filter']


def check_stability(b, a):
    """utility function that checks the stability of the denom and nom of a filter to ensure it's valid

    Parameters
    ----------
    b: Numerator polynomials of the IIR filter
    a: Denominator polynomials of the IIR filter

    Returns
    -------
    None or a warning if unstable filter

    Example Use:
    --------
    b,a = butter(inputs)
    check_stability(b,a)
    """
    z, p, k = tf2zpk(b, a)
    try:
        assert (np.max(np.abs(p)) <= 1)
    except:
        unstable_msg = "Filter is not stable! np.max(np.abs(p)) must be less than 1 but instead is: {} please use sos approach instead"
        warnings.warn(unstable_msg.format(np.max(np.abs(p))), UserWarning)
    return


def apply_correct_filter(filter_inputs, correct_filter, dat, axis):
    """Utility function to apply correct filter to butterworth

    Parameters
    ----------
    filter_inputs: single argument or argument in tuple
    correct_filter: correct filter to apply either filtfilt or sosfiltfilt
    dat: data to apply filter on
    axis: axis on which to apply the filter

    Returns
    -------
    Filtered data
    """
    """utility applies correct filter for sos/ba and unpacks tuple where appropriate"""
    if type(filter_inputs) == tuple:  # * unpacks the tuple
        return correct_filter(*filter_inputs, dat, axis=axis)
    else:
        return correct_filter(filter_inputs, dat, axis=axis)


def butterworth_filter(data=None, freq=(1, 100), filt_type='bandpass',
                       order=4, sample_rate=None, axis=-1,
                       filter_output='sos', linear_filter=False):
    """
    Function for applying a butterworth filter onto data.

    Parameters
    ----------
    data: TimeSeries like
    freq: array-like or int/float depending on filt_type
    filt_type: filter type to use (e.g. 'pass', 'high', 'low', 'highpass',
                'lowpass', 'bandpass', 'band', 'stop', 'bandstop')
    order: int, order of the filter
    sample_rate: int/float, number of samples in a second
    axis: int, axis to filter by default -1
    filter_output: output method for filter, by uses default second order sections
                 valid = ('ba', 'sos')
    linear_filter: bool, NOT IMPLEMENTED YET

    Returns
    -------

    """
    # --------> SANITY CHECKING
    # Make sure the inputted filter type is implemented
    valid_filt_type = ('pass', 'high', 'low', 'highpass', 'lowpass', 'bandpass', 'band', 'stop', 'bandstop')
    if filt_type not in valid_filt_type:
        raise (ValueError('{} was not found in not valid filters\n{}'.format(filt_type, valid_filt_type)))

    # Make sure the high/low pass filters have correct typing
    if (filt_type in ('high', 'low', 'lowpass', 'highpass') and not (type(freq) == float or type(freq) == int)):
        _warning = 'Please use an int or a float for freq input, inputted freq={},type(freq)={}'
        raise (TypeError(_warning.format(freq, type(freq))))

    # If inputted data is a TimeSeries like object, we need to output it as the same type
    is_timeseries = True if issubclass(type(data), TimeSeries) else False
    if is_timeseries:
        # Changing way this is implemented so that it uses less memory...

        #copy = deepcopy(data)
        #data = data.data
        coords = data.coords
        dims = data.dims
        attrs = data.attrs
        name = data.name
        if sample_rate is None:
            sample_rate = float(data['samplerate'].data)

    # Calculate Nyquist frequency to normalize frequency by
    nyquist = sample_rate / 2

    if filt_type in ('low', 'high', 'highpass', 'lowpass'):
        Wn = freq / nyquist
    else:  # All the other can be implemented as two?
        Wn = np.asarray(freq) / nyquist

    # Check which filter we should use, assumes only want second order sections or forward backwards
    correct_filter = sosfiltfilt if filter_output == 'sos' else filtfilt
    if linear_filter:
        # correct_filter = lfilter # Linear phase
        raise NotImplementedError('linear_filt not implemented yet')

    constructed_filter = butter(N=order,  # Order of the filter
                                Wn=Wn,  # Frequency arguments
                                btype=filt_type,  # filter band to use
                                analog=False,  # Digital filter
                                output=filter_output  # ba sos or zpk
                                )
    # Check if we NEED to use second order sections for stability or not
    if filter_output != 'sos':
        check_stability(*constructed_filter)

    # Apply the correct filter and unpacking inputs if necessary
    filtered_data = apply_correct_filter(filter_inputs=constructed_filter,
                                         correct_filter=correct_filter,
                                         dat=np.array(data), axis=axis)
    if is_timeseries:
        return TimeSeriesLF(data=filtered_data, coords=coords, dims=dims, attrs=attrs, name=name)
        #copy.data = filtered_data
        #return copy
    return filtered_data

class ButterworthFilter(BaseFilter):
    """Applies Butterworth filter to a time series.
    Keyword Arguments
    -----------------
    timeseries: TimeSeries Data
         TimeSeries object
    freq_range: list-like
       Array [min_freq, max_freq] describing the filter range
    order: int
         Butterworth filter order
    filt_type: str
         String object in: ('pass', 'high', 'low', 'highpass', 'lowpass',
                            'bandpass', 'band', 'stop', 'bandstop')
    filter_output: str
        String object in: ('sos', 'ba') describing whether to use second order sections or not

    .. versionchanged:: 2.0
        Parameter "time_series" was renamed to "timeseries".


    Notes
    -------
    High-order IIR filters can be highly sensitive to quantization of their coefficients,
    and can easily become unstable. This is much less of a problem with first and second-order
    filters; therefore, higher-order filters are typically implemented as serially-cascaded
    biquad sections (and a first-order filter if necessary). The two poles of the biquad filter
    must be inside the unit circle for it to be stable. In general, this is true for all discrete
    filters i.e. all poles must be inside the unit circle in the Z-domain for the filter to be stable.


    """
    order = traits.api.Int
    # freq_range = traits.api.CList(maxlen=2)
    filt_type = traits.api.Str
    design_valid = ('sos', 'ba', 'zpk')
    valid_filt_type = ('pass', 'high', 'low', 'highpass', 'lowpass',
                       'bandpass', 'band', 'stop', 'bandstop')

    def __init__(self, timeseries, freq_range, order=4, filt_type='stop', filter_output='sos'):
        super(ButterworthFilter, self).__init__(timeseries)

        self.freq_range = freq_range
        self.order = order
        self.filt_type = filt_type
        self.filter_output = filter_output

    def filter(self):
        """
        Applies Butterworth filter to input time series and returns filtered TimeSeries object
        Returns
        -------
        filtered: TimeSeries
            The filtered time series
        """
        time_axis_index = get_axis_index(self.timeseries, axis_name='time')
        filtered_array = butterworth_filter(data=np.array(self.timeseries),
                                            freq=self.freq_range,
                                            filt_type=self.filt_type,
                                            sample_rate=float(self.timeseries['samplerate']),
                                            axis=time_axis_index,
                                            filter_output=self.filter_output,
                                            order=self.order,
                                            linear_filter=False  # DO NOT CHANGE
                                            )
        if issubclass(type(filtered_array), TimeSeries):
            return filtered_array

        coords_dict = {coord_name: DataArray(coord.copy())
                       for coord_name, coord in list(self.timeseries.coords.items())}
        coords_dict['samplerate'] = self.timeseries['samplerate']
        dims = [dim_name for dim_name in self.timeseries.dims]
        filtered_timeseries = TimeSeries(filtered_array, dims=dims, coords=coords_dict)
        filtered_timeseries.attrs = self.timeseries.attrs.copy()
        return filtered_timeseries

def FIR_bandpass_filter(data, lowcut, highcut, fs, order=1000, window='hanning'):
    """Applies a forward reverse FIR bandpass filter according to the parameters
    ----------
    data : TimeSeries or array_like
        The data we would like to apply a bandpass to
    lowcut : float or 1D array_like
        Low cutoff frequency of filter (expressed in the same units as `nyq`)
        The values 0 and `nyq` must not be included in `lowcut`.
    highcut : float or 1D array_like
        High cutoff frequency of filter (expressed in the same units as `nyq`)
        The values 0 and `nyq` must not be included in `highcut`.
    fs : float or 1D array_like
        Sampling frequency of the timeseries

    Order : int or float, by default 1000
        The order of the filter to compute. Increasing the value give better frequency
        resolution of worse temporal resolution
    window : string or tuple of string and parameter values, by default hanning.
        Desired window to use. See `scipy.signal.get_window` for a list
        of windows and required parameters.

    Returns
    -------
    data_filt_ts: TimeSeries
        Data filtered between lowcut and highcut as a TimeSeries object
    data_filt: array_like
        Data filtered between lowcut and highcut as an array_like object

    References
    ----------
    https://gist.github.com/andrewgiessel/7589513
    http://mpastell.com/2010/01/18/fir-with-scipy/
    https://docs.scipy.org/doc/scipy/reference/generated/scipy.signal.filtfilt.html#scipy.signal.filtfilt
    """
    # -----------> Initialize relevant parameters
    n = order + 1  # Length of the filter coefficients, e.g. order + 1
    nyq = 0.5 * fs  # The Nyquist rate of the signal.
    low = lowcut / nyq
    high = highcut / nyq

    # -----------> Construct a Lowpass filter with a window
    lowpass_filt = firwin(n, cutoff=low, window=window)

    # -----------> Construct a Highpass filter with spectral inversion
    highpass_filt = - firwin(n, cutoff=high, window=window)
    highpass_filt[n // 2] = highpass_filt[n // 2] + 1

    # -----------> Combine lowpass and highpass filters into a bandpass filter
    band_pass_filt = - (lowpass_filt + highpass_filt)
    band_pass_filt[n // 2] = band_pass_filt[n // 2] + 1

    # -----------> Compute a forward/reverse bandpass filtering
    data_filt = filtfilt(b=band_pass_filt, a=[1], x=data)

    # If we input a TimeSeries we should output a TimeSeries
    if issubclass(type(data), TimeSeries):
        data_filt_ts = deepcopy(data)
        data_filt_ts.data = data_filt
        return data_filt_ts

    return data_filt


"""inherit_docstrings.py
author = nikratio


class Animal:
    def move_to(self, dest):
        '''Move to *dest*'''
        pass

def check_docstring(fn):
    assert fn.__doc__ == Animal.move_to.__doc__
    return fn

class Bird(Animal, metaclass=InheritableDocstrings):
    @check_docstring
    @copy_ancestor_docstring
    def move_to(self, dest):
        self._fly_to(dest)

assert Animal.move_to.__doc__ == Bird.move_to.__doc__

See Also
http://code.activestate.com/recipes/578587-inherit-method-docstrings-without-breaking-decorat/

"""
from functools import partial


__all__ = ['InheritableDocstrings']

# Replace this with actual implementation from
# http://code.activestate.com/recipes/577748-calculate-the-mro-of-a-class/
# (though this will work for simple cases)
def mro(*bases):
    return bases[0].__mro__


# This definition is only used to assist static code analyzers
def copy_ancestor_docstring(fn):
    '''Copy docstring for method from superclass

    For this decorator to work, the class has to use the `InheritableDocstrings`
    metaclass.
    '''
    raise RuntimeError('Decorator can only be used in classes '
                       'using the `InheritableDocstrings` metaclass')


def _copy_ancestor_docstring(mro, fn):
    '''Decorator to set docstring for *fn* from *mro*'''

    if fn.__doc__ is not None:
        raise RuntimeError('Function already has docstring')

    # Search for docstring in superclass
    for cls in mro:
        super_fn = getattr(cls, fn.__name__, None)
        if super_fn is None:
            continue
        fn.__doc__ = super_fn.__doc__
        break
    else:
        raise RuntimeError("Can't inherit docstring for %s: method does not "
                           "exist in superclass" % fn.__name__)

    return fn


class InheritableDocstrings(type):
    """Object to allow inheriting of docstrings
    class Animal:
        def move_to(self, dest):
            '''Move to *dest*'''
            pass


    class Bird(Animal, metaclass=InheritableDocstrings):
        @copy_ancestor_docstring
        def move_to(self, dest):
            self._fly_to(dest) # Why is this fly to? I get it's a bird but isn't it unresolved?


    assert Animal.move_to.__doc__ == Bird.move_to.__doc__

    # ----> Use with other decorators
    class Animal:
        def move_to(self, dest):
            '''Move to *dest*'''
            pass


    def check_docstring(fn):
        assert fn.__doc__ == Animal.move_to.__doc__
        return fn


    class Bird(Animal, metaclass=InheritableDocstrings):
        @check_docstring
        @copy_ancestor_docstring
        def move_to(self, dest):
            self._fly_to(dest)


    assert Animal.move_to.__doc__ == Bird.move_to.__doc__
    print(Animal.move_to.__doc__, Bird.move_to.__doc__)
    """

    @classmethod
    def __prepare__(cls, name, bases, **kwds):
        classdict = super().__prepare__(name, bases, *kwds)

        # Inject decorators into class namespace
        classdict['copy_ancestor_docstring'] = partial(_copy_ancestor_docstring, mro(*bases))

        return classdict

    def __new__(cls, name, bases, classdict):

        # Decorator may not exist in class dict if the class (metaclass
        # instance) was constructed with an explicit call to `type`.
        # (cf http://bugs.python.org/issue18334)
        if 'copy_ancestor_docstring' in classdict:

            # Make sure that class definition hasn't messed with decorators
            copy_impl = getattr(classdict['copy_ancestor_docstring'], 'func', None)
            if copy_impl is not _copy_ancestor_docstring:
                raise RuntimeError('No copy_ancestor_docstring attribute may be created '
                                   'in classes using the InheritableDocstrings metaclass')

            # Delete decorators from class namespace
            del classdict['copy_ancestor_docstring']

        return super().__new__(cls, name, bases, classdict)


if __name__ == '__main__':
    # Simple use:
    class Animal:
        def move_to(self, dest):
            '''Move to *dest*'''
            pass


    class Bird(Animal, metaclass=InheritableDocstrings):
        @copy_ancestor_docstring
        def move_to(self, dest):
            self._fly_to(dest) # Why is this fly to? I get it's a bird but isn't it unresolved?


    assert Animal.move_to.__doc__ == Bird.move_to.__doc__

    # ----> Use with other decorators
    class Animal:
        def move_to(self, dest):
            '''Move to *dest*'''
            pass


    def check_docstring(fn):
        assert fn.__doc__ == Animal.move_to.__doc__
        return fn


    class Bird(Animal, metaclass=InheritableDocstrings):
        @check_docstring
        @copy_ancestor_docstring
        def move_to(self, dest):
            self._fly_to(dest)


    assert Animal.move_to.__doc__ == Bird.move_to.__doc__
    print(Animal.move_to.__doc__, Bird.move_to.__doc__)




"""loadmat.py
A script used to load .mat files
"""
import scipy.io as spio
import scipy

__all__ = ['loadmat']


MAPPER = {'pickle':'pkl', 'numpy':'npy', 'hdf':'h5', 'csv':'csv', 'edf' : 'edf'}


def loadmat(filename):
    """This function should be called instead of direct scipy.io.loadmat as it cures the problem of not properly
    recovering python dictionaries from mat files.

    Calls the function _check_keys to cure all entries which are still mat-objects

    Parameters
    ----------
    filename: str, filename to a matlab path

    Returns
    -------
    dict: dictionary, data from matlab
    """
    data = scipy.io.loadmat(filename, struct_as_record=False, squeeze_me=True)
    return _check_keys(data)

def _check_keys(dict):
    """checks if entries in dictionary are mat-objects. If yes _todict is called to change them to nested dictionaries

    Parameters
    ----------
    dict: dictionary

    Returns
    -------
    dict: dictionary
    """
    for key in dict:
        if isinstance(dict[key], scipy.io.matlab.mio5_params.mat_struct):
            dict[key] = _todict(dict[key])
    return dict

def _todict(matobj):
    """ A recursive function which constructs from matobjects nested dictionaries

    Parameters
    ----------
    matob

    Returns
    -------

    """
    dict = {}
    for strg in matobj._fieldnames:
        elem = matobj.__dict__[strg]
        if isinstance(elem, scipy.io.matlab.mio5_params.mat_struct):
            dict[strg] = _todict(elem)
        else:
            dict[strg] = elem
    return dict



import numpy as np
from ptsa.data.timeseries import TimeSeries

__all__ = [
    'reorder_dims',
    'xarray_rolling_average',
    'rolling_window_full',
    'rolling_window',
    'rolling_apply'
]
def reorder_dims(darray, dim1, dim2):
    """Interchange two dimensions of a DataArray in a similar way as numpy's swap_axes

    Parameters
    ---------
    darray: Xarray.DataArray/TimeSeriesX
    dim1: str, the first dim we want to swap
    dim2: str, the second dim we want to swap

    Returns
    ---------
    transposed darray

    Example Usage
    -------------
    # Swap the timeseries object's first dim and time dim
    check_first_dim = timeseries.dims[0]
    timeseries_swapped_time_dim = reorder_dims(timeseries, dim1=check_first_dim, dim2='time')

    See
    -------
    https://github.com/pydata/xarray/issues/1081, rafa-guedes comment on Aug 21, 2017
    """
    dims = list(darray.dims)
    assert set([dim1, dim2]).issubset(dims), 'dim1 and dim2 must be existing dimensions in darray'
    ind1, ind2 = dims.index(dim1), dims.index(dim2)
    dims[ind2], dims[ind1] = dims[ind1], dims[ind2]
    return darray.transpose(*dims)


def xarray_rolling_average(timeseries, desired_window, desired_step):
    """Computes a rolling average analogous to sliding_mean_fast using xarray.rolling given a desired window, and step

    Parameters
    ----------
    timeseries: TimeSeriesX, time series we which to average into temporal windows.
                Assumes time dim is explicitly labeled time
    desired_window: float, time in seconds of the desired window (e.g. 500ms/.5s is .5)
    desired_step: float, time in seconds of the desired step between windows (e.g. 10ms is .01)

    Returns
    -------
    timeseries_rolled: TimeSeriesX, time series we inputted but with time dim now representing the midpoint of that time
    point averaged across desired_window seconds.
    """
    first_dim = timeseries.dims[0]
    dims_reordered = False
    if first_dim != 'time':
        # Transpose it so that time is on the first axis for operations
        dims_reordered = True
        timeseries = reorder_dims(darray=timeseries, dim1=first_dim, dim2='time')

    # Use sample rate to determine indexing for windows and steps
    samplerate = 1 / timeseries['samplerate'].data
    timeseries_window_length = int(desired_window / samplerate)
    timeseries_step_length = int(desired_step / samplerate)

    # Using default xarray.DataArray.rolling to compute a mean
    rolling_average = timeseries.rolling(time=timeseries_window_length, center=True).mean()

    # Data is only valid from half the window after start until half before end; e.g. removing nans
    timeseries_rolled = rolling_average[int(timeseries_window_length / 2):int(-timeseries_window_length / 2)]

    # Return windows that correspond with our desired step
    timeseries_rolled = timeseries_rolled[::timeseries_step_length]
    if dims_reordered:
        timeseries_rolled = reorder_dims(darray=timeseries_rolled,
                                         dim1='time',
                                         dim2=first_dim)
    return TimeSeries(data=timeseries_rolled.data, dims=timeseries_rolled.dims, coords=timeseries_rolled.coords)

def rolling_window_full(array, window=(0,), asteps=None, wsteps=None, axes=None, toend=True):
    """Create a view of `array` which for every point gives the n-dimensional
    neighbourhood of size window. New dimensions are added at the end of
    `array` or after the corresponding original dimension.

    Parameters
    ----------
    array : array_like
        Array to which the rolling window is applied.
    window : int or tuple
        Either a single integer to create a window of only the last axis or a
        tuple to create it for the last len(window) axes. 0 can be used as a
        to ignore a dimension in the window.
    asteps : tuple
        Aligned at the last axis, new steps for the original array, ie. for
        creation of non-overlapping windows. (Equivalent to slicing result)
    wsteps : int or tuple (same size as window)
        steps for the added window dimensions. These can be 0 to repeat values
        along the axis.
    axes: int or tuple
        If given, must have the same size as window. In this case window is
        interpreted as the size in the dimension given by axes. IE. a window
        of (2, 1) is equivalent to window=2 and axis=-2.
    toend : bool
        If False, the new dimensions are right after the corresponding original
        dimension, instead of at the end of the array. Adding the new axes at the
        end makes it easier to get the neighborhood, however toend=False will give
        a more intuitive result if you view the whole array.

    Returns
    -------
    A view on `array` which is smaller to fit the windows and has windows added
    dimensions (0s not counting), ie. every point of `array` is an array of size
    window.

    Examples
    --------
    >>> a = np.arange(9).reshape(3,3)
    >>> rolling_window(a, (2,2))
    array([[[[0, 1],
             [3, 4]],
            [[1, 2],
             [4, 5]]],
           [[[3, 4],
             [6, 7]],
            [[4, 5],
             [7, 8]]]])

    Or to create non-overlapping windows, but only along the first dimension:
    >>> rolling_window(a, (2,0), asteps=(2,1))
    array([[[0, 3],
            [1, 4],
            [2, 5]]])

    Note that the 0 is discared, so that the output dimension is 3:
    >>> rolling_window(a, (2,0), asteps=(2,1)).shape
    (1, 3, 2)

    This is useful for example to calculate the maximum in all (overlapping)
    2x2 submatrixes:
    >>> rolling_window(a, (2,2)).max((2,3))
    array([[4, 5],
           [7, 8]])

    Or delay embedding (3D embedding with delay 2):
    >>> x = np.arange(10)
    >>> rolling_window(x, 3, wsteps=2)
    array([[0, 2, 4],
           [1, 3, 5],
           [2, 4, 6],
           [3, 5, 7],
           [4, 6, 8],
           [5, 7, 9]])


    see:
    https://gist.github.com/seberg/3866040
    """
    array = np.asarray(array)
    orig_shape = np.asarray(array.shape)
    window = np.atleast_1d(window).astype(int)  # maybe crude to cast to int...

    if axes is not None:
        axes = np.atleast_1d(axes)
        w = np.zeros(array.ndim, dtype=int)
        for axis, size in zip(axes, window):
            w[axis] = size
        window = w

    # Check if window is legal:
    if window.ndim > 1:
        raise ValueError("`window` must be one-dimensional.")
    if np.any(window < 0):
        raise ValueError("All elements of `window` must be larger then 1.")
    if len(array.shape) < len(window):
        raise ValueError("`window` length must be less or equal `array` dimension.")

    _asteps = np.ones_like(orig_shape)
    if asteps is not None:
        asteps = np.atleast_1d(asteps)
        if asteps.ndim != 1:
            raise ValueError("`asteps` must be either a scalar or one dimensional.")
        if len(asteps) > array.ndim:
            raise ValueError("`asteps` cannot be longer then the `array` dimension.")
        # does not enforce alignment, so that steps can be same as window too.
        _asteps[-len(asteps):] = asteps

        if np.any(asteps < 1):
            raise ValueError("All elements of `asteps` must be larger then 1.")
    asteps = _asteps

    _wsteps = np.ones_like(window)
    if wsteps is not None:
        wsteps = np.atleast_1d(wsteps)
        if wsteps.shape != window.shape:
            raise ValueError("`wsteps` must have the same shape as `window`.")
        if np.any(wsteps < 0):
            raise ValueError("All elements of `wsteps` must be larger then 0.")

        _wsteps[:] = wsteps
        _wsteps[window == 0] = 1  # make sure that steps are 1 for non-existing dims.
    wsteps = _wsteps

    # Check that the window would not be larger then the original:
    if np.any(orig_shape[-len(window):] < window * wsteps):
        raise ValueError("`window` * `wsteps` larger then `array` in at least one dimension.")

    new_shape = orig_shape  # just renaming...

    # For calculating the new shape 0s must act like 1s:
    _window = window.copy()
    _window[_window == 0] = 1

    new_shape[-len(window):] += wsteps - _window * wsteps
    new_shape = (new_shape + asteps - 1) // asteps
    # make sure the new_shape is at least 1 in any "old" dimension (ie. steps
    # is (too) large, but we do not care.
    new_shape[new_shape < 1] = 1
    shape = new_shape

    strides = np.asarray(array.strides)
    strides *= asteps
    new_strides = array.strides[-len(window):] * wsteps

    # The full new shape and strides:
    if toend:
        new_shape = np.concatenate((shape, window))
        new_strides = np.concatenate((strides, new_strides))
    else:
        _ = np.zeros_like(shape)
        _[-len(window):] = window
        _window = _.copy()
        _[-len(window):] = new_strides
        _new_strides = _

        new_shape = np.zeros(len(shape) * 2, dtype=int)
        new_strides = np.zeros(len(shape) * 2, dtype=int)

        new_shape[::2] = shape
        new_strides[::2] = strides
        new_shape[1::2] = _window
        new_strides[1::2] = _new_strides

    new_strides = new_strides[new_shape != 0]
    new_shape = new_shape[new_shape != 0]

    return np.lib.stride_tricks.as_strided(array, shape=new_shape, strides=new_strides)

def rolling_apply(fun, a, w):
    """

    Parameters
    ----------
    fun
    a
    w

    Returns
    -------

    """
    # http://www.rigtorp.se/2011/01/01/rolling-statistics-numpy.html
    r = np.empty(a.shape)
    r.fill(np.nan)
    for i in range(w - 1, a.shape[0]):
        r[i] = fun(a[(i - w + 1):i + 1])
    return r


def rolling_window(a, window):
    """

    Parameters
    ----------
    a
    window

    Returns
    -------

    """
    # http://www.rigtorp.se/2011/01/01/rolling-statistics-numpy.html
    shape = a.shape[:-1] + (a.shape[-1] - window + 1, window)
    strides = a.strides + (a.strides[-1],)
    return np.lib.stride_tricks.as_strided(a, shape=shape, strides=strides)


"""staging.py

The purpose of this script is allow one to automatically classify intracranial eeg data as either wake, Non-REM,
or not Non-REM. The way we calculate the stage is through the use of a metric called Visualization Index (VI).

Steps:
Window data into 30s epochs with 0s of overlap
Run a FFT over each 30s epoch
Normalize the resulting PSD across time and channels to 1 using the total area under the power spectrum.
Across subjects: normalize the scale so maximal VI value, for each channel, is equal to 100 and minimum is 1
Average all channels

Definitions:
                        delta = (1,4)
                        theta = (4, 8)
                        alpha = (8, 12)
                        spindle = (11,16)
                        high_beta = (20, 40)
                        VI = (delta + theta + spindle) / (high_beta, alpha)
"""
import numpy as np
from cmlreaders import CMLReader
from Clumsy import get_sub_tal
#from .io import get_sub_tal
from glob import glob
import os
from ptsa.data.filters import MonopolarToBipolarMapper
from IPython.display import display
from Clumsy import psd, rolling_window_full, TimeSeriesLF
from .data_organization import get_data_roots, get_subj_exp_sess_date_matcher
SUBJ_EXP_SESS_DATE_MATCHER = get_subj_exp_sess_date_matcher()
DATA_ROOTS = get_data_roots()

__all__ = ['VigilanceIndexPipeline']

# TODO: add more helpful warnings in, e.g. instead of keyerror from SUBJ_EXP do session not valid, if no paths exist print that.
class Error(BaseException):
    """Base class for exceptions in this module."""
    def __init__(self, *args, **kwargs):
        BaseException.__init__(self, *args, **kwargs)

class DoneGoofed(Error):
    """Exception raised when no paths are found
    """
    def __init__(self, warning='', *args, **kwargs):
        Error.__init__(self, warning)

class DoneGoofed_InvalidSession(Error):
    """Exception raised for errors in the input, here if they didn't enter a valid session.
    Attributes:
        session -- input session of thes
        possible_sessions -- valid sessions of the subject
    """

    def __init__(self, session, possible_sessions):
        warning = 'DoneGoofed Session Error: Ah Shucks sorry to say but it looks someone done goofed...'
        warning2 = 'I could not find session {}. Have you considered using a valid session instead? Try: \n'
        self.warning = warning2.format(session)
        print(warning)
        print(self.warning)
        display(possible_sessions)

class VigilanceIndexPipeline(object):
    """Calculates the Vigilance index, a measure of depth of NREM _sleep

    Parameters
    ----------
    subject: str,
             subject ID, e.g. 'R1293P'
    experiment: str,
                experiment analyze
    session: str/int,
             behavioral session to analyze
    montage: str, by default 'bipolar',
             montage scheme to apply,
             valid: 'bipolar', 'bi', 'bp', 'pairs', 'monopolar', 'mono', 'mp', 'contacts'
    window_len: int, by default 30,
                length in seconds of window to calculate index over
    overlap: int, by default 0,
             length in seconds to allow overlap of windows when calculating index
    number_lat_chs: int, by default 10,
                    number of channels to include sorted by lateralization in the brain
    night_start_time: str, by default '2200',
                      time in 2400 to consider a night to start at, used in night before calculations
    morning_start_time: str, by default '0800',
                        time in 2400 to consider a morning to began at, used for night after calculations

    Notes
    -------
    The power spectrum density is normalized so that the total area under the curve is 1, afterwards this value must be
    normalized between 1-100 if comparing across subjects. By default this object will conduct both normalizations

    The purpose of this object is allow one to automatically classify intracranial eeg data as either Non-REM,
    or not Non-REM. The way we calculate the stage is through the use of a metric called Visualization Index (VI).

    Steps:
    Window data into 30s epochs with 0s of overlap
    Run a FFT over each 30s epoch
    Normalize the resulting PSD across time and channels to 1 using the total area under the power spectrum.
    Across subjects: normalize the scale so maximal VI value, for each channel, is equal to 100 and minimum is 1
    Average all channels

    Definitions:
                            delta = (1,4)
                            theta = (4, 8)
                            alpha = (8, 12)
                            spindle = (11,16)
                            high_beta = (20, 40)
                            VI = (delta + theta + spindle) / (high_beta, alpha)


    """
    VI_frequency_bands = {
        'delta' : (1, 4),
        'theta' :(4, 8),
        'alpha' : (8, 12),
        'spindle' : (11, 16),
        'high_beta' : (20, 40),
    }
    delta = (1, 4)
    theta = (4, 8)
    alpha = (8, 12)
    spindle = (11, 16)
    high_beta = (20, 40)
    VI = '(delta + theta + spindle) / (high_beta, alpha)'

    hpc_labels = np.array(['Left CA1', 'Left CA2', 'Left CA3', 'Left DG', 'Left Sub',
                           'Right CA1', 'Right CA2', 'Right CA3', 'Right DG', 'Right Sub'])
    mtl_labels = np.append(hpc_labels, np.array(['Left PRC', 'Left EC', 'Right PRC', 'Right EC']))
    downsampled_root = '/data/eeg/LFSleep/ProcData/{}/**/*100hz.h5'
    fmt = "%m_%d_%y_%H%M" # month, day, year (decimal format, e.g. 18 not 2018), hour in 2400, minute

    def __init__(self, subject, experiment, session, montage='bipolar', window_len=30, overlap=0, number_lat_chs=10,
                 night_before=True, night_start_time='2200', morning_start_time='0800'):
        """Calculates the Vigilance index, a measure of depth of NREM _sleep
           Where:
           delta = (1, 4)
           theta = (4, 8)
           alpha = (8, 12)
           spindle = (11, 16)
           high_beta = (20, 40)
           VI = (delta + theta + spindle) / (high_beta, alpha)


        Parameters
        ----------
        subject: str,
                 subject ID, e.g. 'R1293P'
        experiment: str,
                    experiment analyze
        session: str/int,
                 behavioral session to analyze
        montage: str, by default 'bipolar',
                 montage scheme to apply,
                 valid: 'bipolar', 'bi', 'bp', 'pairs', 'monopolar', 'mono', 'mp', 'contacts'
        window_len: int, by default 30,
                    length in seconds of window to calculate index over
        overlap: int, by default 0,
                 length in seconds to allow overlap of windows when calculating index
        number_lat_chs: int, by default 10,
                        number of channels to include sorted by lateralization in the brain
        night_before: bool, by default True,
                      whether to look at the night before (True) or after (False) the experiment
        night_start_time: str, by default '2200',
                          time in 2400 to consider a night to start at, used in night before calculations
        morning_start_time: str, by default '0800',
                            time in 2400 to consider a morning to began at, used for night after calculations
        """
        self.subject = subject
        self.experiment = experiment
        self.session = session
        self.montage = montage
        self.window_len = window_len # In seconds
        self.overlap = overlap # In seconds
        self.n = number_lat_chs # Calculate using n most lateral channels
        self.night_before = night_before
        self.night_start_time = night_start_time
        self.morning_start_time = morning_start_time

        self.mp, self.bp, self.tal = get_sub_tal(subject, experiment, True)
        self.reader = CMLReader(subject=self.subject,
                                experiment=self.experiment,
                                session=self.session)
        self.pairs = self.reader.load('pairs')
        self.contacts = self.reader.load('contacts')
        self.electrode_categories = self.reader.load('electrode_categories')
        self.epileptic_channels = np.unique(np.concatenate((self.electrode_categories['soz'],
                                                            self.electrode_categories['interictal'])))

        if self.montage in ('bipolar', 'bi', 'bp', 'pairs'):
            self.channels_df = self.pairs
            self.montage = 'bipolar'
        elif self.montage in ('monopolar', 'mono', 'mp', 'contacts'):
            self.channels_df = self.contacts
            self.montage = 'monopolar'
        else:
            raise ValueError(
                'Expected either "monopolar" or "bipolar" as montage input but got {}'.format(self.montage))

        self.sleep_paths = glob(VigilanceIndexPipeline.downsampled_root.format(self.subject), recursive=True)
        try:
            self.experimental_date = SUBJ_EXP_SESS_DATE_MATCHER[self.subject.upper(
            )][self.experiment.lower()][str(self.session)]
        except KeyError: # If session is invalid
            df = CMLReader.get_data_index()
            df = df[df['subject'] == self.subject]
            DoneGoofed_InvalidSession(session=self.session, possible_sessions=df[['experiment', 'session']].T)

        self.load_data_paths = self.get_sleep_recording_paths()

        return

    def run_detection(self):
        """Main function to call to run the pipeline

        Returns
        -------
        Vigilance Index, normalized between values of 0-100 where higher numbers indicate deeper _sleep
        """
        ts = self.fft_from_whole_night()

        # Normalize by area under the curve (values sum to 1)
        ts = self._normalize_by_integral(ts)
        #ts = self.average_frequencies_into_bands(ts)

        ts = self.frequencies_into_bands(ts, how = 'sum')
        VI = (ts.sel(frequency='delta') + ts.sel(frequency='theta') + ts.sel(frequency='spindle')
             ) / (ts.sel(frequency='alpha') + ts.sel(frequency='high_beta'))
        return self._normalize_zero_to_high(VI.mean('channels'))


    def sort_channels_by_lateralization(self):
        """Returns an array of the most lateral channels in the subject

        Returns
        -------
        channels_sorted_laterally: np.array, indices corresponding to the sorted channels
        """
        absolute_lateral = np.abs(np.array(self.channels_df['ind.x']))  # np.abs ensures handling left/right the same
        channels_sorted_laterally = np.argsort(absolute_lateral)[::-1]
        return channels_sorted_laterally

    def get_n_most_lateral_channel_indices(self, n=None):
        """Return indices corresponding to n most lateral channels

        Parameters
        ----------
        n: int, by default None
           the number of channels to include, only change this when it's desirable to not use the instance's attribute

        Returns
        -------
        lateral_channel_indices: np.array, indices corresponding to the desired channels
        """
        if n is None:
            n = self.n

        if self.montage == 'monopolar':
            # Get n most lateral contacts that are not epileptic
            keep = ~np.isin(self.sort_channels_by_lateralization(),
                            np.where(np.isin(self.channels_df['label'], self.epileptic_channels))[0])
            lateral_channel_indices = self.sort_channels_by_lateralization()[keep][:n]
            return lateral_channel_indices

        elif self.montage == 'bipolar':
            ch0= np.array(list(map(lambda s:s.split('-')[0], self.channels_df['label'])))
            ch1 = np.array(list(map(lambda s:s.split('-')[1], self.channels_df['label'])))
            keep0 = ~np.isin(self.sort_channels_by_lateralization(),
                             np.where(np.isin(ch0, self.epileptic_channels))[0])
            keep1 = ~np.isin(self.sort_channels_by_lateralization(),
                             np.where(np.isin(ch1, self.epileptic_channels))[0])
            lateral_channel_indices = self.sort_channels_by_lateralization()[keep0 & keep1][:n]
            return lateral_channel_indices

    def filter_lateral_channels(self, timeseries):
        """Fetch relevant lateral channels and filters the data using montage scheme

        Parameters
        ----------
        timeseries: TimeSeriesLF, MONOPOLAR timeseries to filter

        Returns
        -------
        timeseries: TimeSeriesLF, timeseries containing only the lateral channels, either as monopolar referrenced or
        bipolar referenced depending on self.montage
        """
        relevant_channels = self.get_n_most_lateral_channel_indices()
        # Make sure we only pick valid channels
        timeseries = timeseries.sel(channels=np.isin(timeseries.channels, self.contacts['label']))
        if self.montage == 'monopolar':
            timeseries = timeseries[relevant_channels]

        # Need to re-reference from monopolar data in the path
        elif self.montage == 'bipolar':
            valid_mp = np.unique(np.concatenate((self.bp['ch0'][relevant_channels],
                                                 self.bp['ch1'][relevant_channels])))
            valid_bp = self.bp[relevant_channels]
            timeseries['channels'].data = self.mp
            timeseries = timeseries.sel(channels=valid_mp)
            mono2bi = MonopolarToBipolarMapper(timeseries, bipolar_pairs=valid_bp)
            timeseries = mono2bi.filter()
        return timeseries

    def get_sleep_recording_paths(self):
        """Return an array of paths for all valid 100hz resampled data for a given session's previous night or
        subsequent night

        Returns
        -------
        path_files: np.array, array of valid paths from the night before the relevant session
        """
        # Path formatting is month_day_year_start_stop, e.g. '8_12_11_0320_0450... .h5'
        dates_formatted = ['_'.join(x.split('_')[:4]) for x in
                           list(map(os.path.basename, self.sleep_paths))]

        # Date the experiment took place and relevant datetime string formatting
        start_date_formatted = '_'.join(self.experimental_date.split('_')[:4])
        month, day, year, hour_minutes = start_date_formatted.split('_')
        night_before_start_date_formatted = '_'.join([month, str(int(day) - 1), year, self.night_start_time])
        morning_after_start_date_formatted = '_'.join([month, str(int(day) + 1), year, self.morning_start_time])

        # Go through all paths subject has on rhino and pull any that match signature pattern
        path_files = []
        for i, date_path in enumerate(dates_formatted):

            if self.night_before:
                if (date_path < start_date_formatted) and (date_path > night_before_start_date_formatted):
                    path_files.append(self.sleep_paths[i])

            elif not self.night_before:
                if (date_path > start_date_formatted) and (date_path < morning_after_start_date_formatted):
                    path_files.append(self.sleep_paths[i])
        if len(path_files) == 0:
            raise DoneGoofed('No paths found, ensure that the data are properly uploaded to rhino in the format requested')
        return sorted(np.array(path_files))


    def load_whole_night_rolled_intervals(self):
        """Load the entire night's downsampled data, chunking into self.window_len long intervals

        Returns
        -------
        rolled_data: np.array, data rolled into 30s chunks concatenated across paths/epochs
        """
        # Create rolling windows to compute power spectral density over
        rolling = []
        for path in self.load_data_paths:
            dat = TimeSeriesLF.from_hdf(path)
            dat = self.filter_lateral_channels(dat)

            step = dat._TimeSeries__duration_to_samples(self.window_len)
            asteps = dat._TimeSeries__duration_to_samples(self.window_len - self.overlap)
            arr = dat.data
            rolled = rolling_window_full(array=arr, window=step, asteps=asteps)
            rolling.append(rolled.transpose(1, 0, 2)) # Transpose so we can concat over paths/epochs
        try:
            rolled_data = np.concatenate(rolling).transpose(1, 0, 2)
        except ValueError as e:
            # If there aren't any files
            print(e)

        return rolled_data

    def fft_from_whole_night(self):
        """Calculates rolling Fast Fourier Transformation using Welch's mean method


        Returns
        -------
        ts: TimeSeriesLF,
            time series data of power spectrum density across a whole night's recording
        """
        # Create rolling windows to compute power spectral density over
        rolled_data = self.load_whole_night_rolled_intervals()
        channels = self.channels_df['label'][self.get_n_most_lateral_channel_indices()]
        fs = 100
        frequencies, data = psd(rolled_data, Fs=fs)
        # Make it into a time series
        coords = {'channels': channels,
                  'epochs': np.arange(data.shape[1]),
                  'frequency': frequencies,
                  'samplerate': fs}
        dims = ['channels', 'epochs', 'frequency']
        ts = TimeSeriesLF(data, dims=dims, coords=coords)
        ts = ts[:,:,1:41] # Only need frequencies between 1:40Hz
        return ts


    ########### NORMALIZATION STATIC METHODS ###########
    @staticmethod
    def _normalize_min_to_max(array, minimum, maximum):
        # Normalize channels to between .0 and +1.0
        return (array - minimum) / (maximum - minimum)

    @staticmethod
    def _normalize_zero_to_high(data, high=100, axis=-1):
        """ Normalize data to values between 0 and high

        Parameters
        ----------
        data: np.array,
              data to normalize
        high: int, by default 100,
              value to set as the relative maximum
        axis: int, by default -1,
              axis over which to apply normalization

        Returns
        -------
        inputted data normalized between 0 and high
        """
        c_ = np.array(data)
        arr = c_.T * np.array([high]) / np.max(np.abs(c_), axis=axis)
        return arr.T

    @staticmethod
    def _normalize_by_integral(timeseries, dx=1., axis=1):
        """Apply an integral to normalize a given timeseries

        Parameters
        ----------
        timeseries: TimeSeriesLF,
                    time series data to normalize
        dx: float, default=1.,
            step over which to integrate on y
        axis: int, by default -1,
            axis over which to integrate

        Returns
        -------
        normalized: TimeSeriesLF,
                    time series data normalized
        """
        integral = VigilanceIndexPipeline._integrate(timeseries, dx=dx, axis=axis)
        # Norm so each ch sums to 1
        # TODO: this could be done more intelligently.
        if axis == 0:
            normalized = timeseries / integral[np.newaxis, :, :]
        elif axis == 1:
            normalized = timeseries / integral[:, np.newaxis, :]
        elif axis == 2:
            normalized = timeseries / integral[:, :, np.newaxis]
        return normalized

    ########### GENERAL UTILITY STATIC METHODS ###########

    @staticmethod
    def _integrate(y, dx=1., axis=1):
        """Calculates the discrete integral along the given axis using the composite trapezoidal rule

        Parameters
        ----------
        y: array-like
            input array for integral
        dx: float, default=1.,
            step over which to integrate on y
        axis: int, by default -1,
            axis over which to integrate

        Returns
        -------
        integral: array-like
                  shape = (y.shape[:-1])
        """
        integral = np.trapz(y=y, dx=dx, axis=axis)
        return integral

    @staticmethod
    def valid_band(freq, low, high):
        """Given an array of frequencies return a boolean array that is True bounded between high and low

        Parameters
        ----------
        freq: np.array like,
              array of frequency values
        low: int,
             lower bound of desired frequency range
        high: int,
              upper bound of desired frequency range

        Returns
        -------
        boolean_array of valid boundaries
        """
        return (freq >= low) & (freq <= high)

    @staticmethod
    def frequencies_into_bands(timeseries, how='sum'):
        """Convert frequencies into bands using either the frequencies mean or sum

        Parameters
        ----------
        timeseries: TimeSeriesLF
        how: str, by default 'sum',
             valid = 'sum', 'mean'

        Returns
        -------
        dat: TimeSeriesLF, data averaged into bands
        """
        frequency_bands = VigilanceIndexPipeline.VI_frequency_bands
        dat = []
        for k,v in enumerate(frequency_bands):
            _freqs = VigilanceIndexPipeline.valid_band(timeseries.frequency.data, *frequency_bands[v])
            if how == 'sum':
                dat.append(timeseries.sel(frequency=_freqs).sum('frequency'))
            elif how == 'mean':
                dat.append(timeseries.sel(frequency=_freqs).mean('frequency'))
        dat = TimeSeriesLF.concat(dat, dim='frequency')
        dat['frequency'] = list(frequency_bands.keys())
        return dat


"""detect.py Module for detection of ripples, spindles, and inter-ictal discharges"""
from ptsa.data.timeseries import TimeSeries
from ptsa.data.common import get_axis_index
from scipy.stats import zscore
from Clumsy import rolling_window_full, ButterworthFilter
from numba import jit
import traits.api
import numpy as np
import pandas as pd
#from .base import BaseDetector, jit_find_containing_intervals, find_consecutive_data
from Clumsy import HilbertFilter
from .detection_utils import jit_find_containing_intervals, find_consecutive_data
from .base import BaseDetector
__all__ = ['SpindleDetector',
           'IEDDetector',
           'RippleDetector',
           'BaseDetector',
           #'jit_find_containing_intervals',
           #'find_consecutive_data'
           ]

# --------------> Event Detection Objects

class SpindleDetector(BaseDetector):
    """
    """
    # Spindle definitions
    min_sample, max_sample = .5, 3.
    min_freq, max_freq = 11., 16.
    #duration = (.5, 3)

    methods = {'rms': { 'window' : .2 ,
                        'asteps' : None,
                        'freq' : (11., 16),
                        'threshold' : 92
                      }
               }

    def __init__(self, time_series, event_type='spindle', method='rms', duration=(.5, 3)):


        super(BaseDetector, self).__init__(time_series=time_series, event_type=event_type, method=method)

        self.min_sample = self.time_series._TimeSeries__duration_to_samples(duration[0])
        self.max_sample = self.time_series._TimeSeries__duration_to_samples(duration[1])

        pass

    def rms(self):
        self.rootmeansquare(window=.2, asteps=None, dim='time')

    @staticmethod
    def find_intervals(boolean_arr, min_sample, max_sample, contacts_df=None):
        """Find intervals where there is an amplitude and duration crossing
        ------
        INPUTS:
        ------
        ts: TimeSeriesLF object

        ------
        OUTPUTS:
        ------
        data: list, [ch x array(n x 2)] of start, stop indicies
        """
        try:
            fs_roi = contacts_df['ind.region']
            stein_roi = contacts_df['stein.region']
            hemi = ['left' if x == -1 else 'right' for x in np.sign(contacts_df['ind.x'])]
            channels = contacts_df['label']
        except AttributeError:
            fs_roi = ''
            stein_roi = ''
            hemi = ''
            channels = ''
        except Exception as e:
            print(e)
            fs_roi = ''
            stein_roi = ''
            hemi = ''
            channels = ''

        dataframe = []
        for index, ch_arr in enumerate(boolean_arr):
            start, stops = jit_find_containing_intervals(ch_arr, min_sample)
            df = pd.DataFrame(start, columns=['start'])
            try:
                df['stop'] = stops
            except ValueError as e:
                if len(start) != len(stops):  # Should occur when none are found
                    continue
                print(e)
            df['duration'] = df['stop'] - df['start']
            df['channel'] = channels[index]
            df['fs region'] = fs_roi[index]
            df['hemisphere'] = hemi[index]
            df['stein region'] = stein_roi[index]
            df = df[df['duration'] < max_sample]
            dataframe.append(df)
        try:
            return pd.concat(dataframe)
        except ValueError as e:
            print(e)
            return dataframe


class IEDDetector(BaseDetector):
    """
    """
    # IED definitions

    methods = {'buzsaki': {'thres_filtered': 2,
                           'thres_unfiltered': 2,
                           'freq_range': (25, 80)}

               }

    def __init__(self, time_series, event_type='ied', method='buzsaki'):
        super(BaseDetector, self).__init__(time_series=time_series, event_type=event_type, method=method)
        self.method_d = IEDDetector.methods[method]
        self.frequency = self.method_d['freq_range']
        return

    def detect_ied(self):
        ts = self.remove_linenoise(self.time_series, harmonic=2, line_noise=60)
        ied = ts.filter_with(ButterworthFilter, freq_range=self.frequency,
                             order=4, filt_type='pass', filter_output='sos')

        thres_filtered = self.method_d['thres_filtered']
        thres_unfiltered = self.method_d['thres_unfiltered']

        # More than n times filtered data -> Keep
        """Keep any time points that exceed the threshold of the filtered data"""
        mean = np.mean(ied.data, 1)
        std = np.std(ied.data, 1)
        thres = mean + (thres_filtered * std)
        thres = thres[:, None]
        boolean_array = ied.data > thres

        # Less than n times baseline -> Remove
        """Discard any time points that are below the threshold of the unfiltered data"""
        mean_raw = np.abs(ts).mean('time')
        std_raw = np.abs(ts).std('time')
        thres_raw = mean_raw.data + (thres_unfiltered * std_raw.data)
        thres_raw = thres_raw[:, None]
        bool_raw = self.time_series.data > thres_raw

        return (bool_raw & boolean_array)

    def detect(self):
        boolean_arr = self.detect_ied()
        dfs = []
        for i, boolean_ch in enumerate(boolean_arr):
            consec = find_consecutive_data(np.where(boolean_ch)[0])
            try:
                consec = np.array([x[0] for x in consec])
            except IndexError as e:
                continue

            consec = consec[np.append(True, np.diff(consec) > self._duration_to_samples(duration=1.))]
            df = pd.DataFrame(consec, columns=['samples'])
            df['channel'] = str(self.time_series.channels[i].data)
            df['event'] = 'IED'
            dfs.append(df)
        try:
            return pd.concat(dfs)
        except ValueError as e:
            return dfs

class RippleDetector(BaseDetector):

    #def __init__(self, time_series):
        #super(RippleDetector, self).__init__(time_series)

    methods = {'buzsaki': {'method': 'zscore',
                           'value threshold': 3.,
                           'freq_range': (80, 250),
                           'duration threshold':.2
                           }

               }

    def __init__(self, time_series, event_type='ripple', method='buzsaki'):
        # TODO: Use *args super()
        super(BaseDetector, self).__init__(time_series=time_series, event_type=event_type, method=method)
        self.method_d = RippleDetector.methods[method]
        self.frequency = self.method_d['freq_range']
        self.order = 4

        return

    @staticmethod
    def _ripple_filter_amplitude_envelope(ts, freq_range=(80,250), order=4):
        """

        Parameters
        ----------
        ts
        freq_range
        order

        Returns
        -------

        """
        # Filter to ripple
        ripple_filt = ts.filter_with(ButterworthFilter, freq_range=freq_range, order=order,
                                     filt_type='pass', filter_output='sos')
        ripple_filt = ripple_filt.filter_with(HilbertFilter)
        ripple_filt.data = np.abs(ripple_filt)
        return ripple_filt

    def ripple_filter_amplitude_envelope(self):
        return  self._ripple_filter_amplitude_envelope(ts=self.time_series, freq_range=self.frequency, order=self.order)

    def detect(self, ripple_filt, contacts):
        from Clumsy import chop_intervals, gaussian_smooth, RippleDetector, SpindleDetector
        ripple_filt = self.ripple_filter_amplitude_envelope()
        ripple_filt = gaussian_smooth(ripple_filt)
        boolean_arr = RippleDetector.zscore_threshold(ripple_filt, threshold=3, axis=-1)
        min_sample, max_sample = .2, None
        min_sample = ripple_filt._TimeSeries__duration_to_samples(min_sample)
        df = BaseDetector.find_intervals(boolean_arr, min_sample, max_sample, contacts)


"""
Staresina_2018 -> Hierarchical nesting of slow oscillations, spindles and ripples in the human hippocampus during _sleep

https://www.nature.com/articles/nn.4119.pdf

Event detection and extraction. SO, spindle and ripple events were identified independently for each participant and 
channel based on established detection algorithms15,23. 

SOs were detected as follows. First, data were filtered between 0.16–1.25 Hz (two-pass FIR bandpass filter, order = 3 
cycles of the low frequency cut-off), and only  artifact-free data from NREM _sleep stages 2–4 were used for event detection. 
Second, all zero-crossings were  determined in the filtered signal, and event duration was determined for SO candidates 
(that is, down-states followed by up-states) as time between two successive positive- to-negative zero-crossings for Cz 
and two successive negative-to-positive zero-crossings for HC, respectively. Events that met the SO duration criteria 
(minimum of 0.8 and maximum of 2 s, 0.5–1.25 Hz) entered the next step. Third, event amplitudes were determined 
for the remaining SO candidates (trough-to-peak amplitude between two positive-to-negative zero crossing for Cz; 
peak-to-trough amplitude between two negative-to-positive zero-crossing for HC). Events that also met the SO amplitude 
criteria (≥75% percentile of SO candidate amplitudes, that is, the 25% of events with the largest amplitudes) were 
considered SOs. Manual validation in a random sampling of the raw EEG data yielded good agreement between hand-scored 
and algorithmically identified SOs, however with greater sensitivity of the automated algorithm for SOs that were less 
pronounced against the background EEG activity. Finally, artifact-free epochs (−2.5 to +2.5 s) time-locked to the SO
down-state in the filtered signal were extracted from the unfiltered raw signal for all events.

Spindles were detected as follows. First, data were filtered between 12–16 Hz (two-pass FIR bandpass filter, order = 3 
cycles of the low frequency cut-off), and only artifact-free data from NREM _sleep stages 2–4 were used for event 
detection. Second, the r.m.s. signal was calculated for the filtered signal using a moving average of 200 ms, and the 
spindle amplitude criterion was defined as the 75% percentile of RMS values. Third, Whenever the signal exceeded this 
threshold for more than 0.5 s but less than 3 s (duration criteria) a spindle event was detected. Again, manual 
validation in a random sampling of the raw EEG data yielded good agreement between hand-scored and algorithmically 
identified spindles, however with greater sensitivity of the automated algorithm for spindles that were less 
pronounced against the background EEG activity. Finally, artifact-free epochs (−2.5 to +2.5 s) time-locked to the 
maximum spindle trough in the filtered signal were extracted from the unfiltered raw signal for all events.


Ripples were detected as follows. First, data were filtered between 80–100 Hz (two-pass FIR bandpass filter, order = 3 
cycles of the low frequency cut-off), and only artifact-free data from NREM _sleep stages 2–4 were used for event 
detection. Second, the r.m.s. signal was calculated for the filtered signal using a moving average of 20 ms, and the 
ripple amplitude criterion was defined as the 99% percentile of RMS values. Third, whenever the signal exceeded this 
threshold for a minimum of 38 ms (encompassing ~3 cycles at 80 Hz) a ripple event was detected. In addition, we 
required at least three discrete peaks or three discrete troughs to occur in the raw signal segment corresponding to 
the above-threshold RMS segment. This was accomplished by identifying local maxima or minima in the respective raw 
signal segments after applying a one-pass moving average filter
"""


# Pizza ptsa stuff
from ptsa.data.timeseries import TimeSeries
from ptsa.data.filters import ButterworthFilter
from pyedflib import EdfReader, EdfWriter
import pyedflib
import numpy as np

__all__ = ['preprocess_raw_edf']

def preprocess_raw_edf(file_path, correct_channels, save_path):
    """Given an edf filepath this will use partial string matches to remove any not matched channels (e.g DC channels etc) and notch filter
    INPUTS
    ------
    file_path: str, path to edf
    correct_channels: np.array, an array of the correct channels for the subject, created using CMLReaders
        e.g. :
        reader = CMLReader(subject='R1207J', experiment="FR1", session=1,
                           localization=0, montage=0)
        ch_df = reader.load('contacts') # Monopolar channels
        correct_channels = np.array(ch_df['label']).view(np.recarray)
    save_path: str, location to save at
    """
    print('Starting to making file: ', save_path)
    f = EdfReader(file_name=file_path)
    n = f.signals_in_file
    signal_labels = f.getSignalLabels()

    # Find partial matches in edf raw signals with cmlreaders channel labels
    signal_labels = np.array([label.upper() for label in signal_labels])
    match_indx = []
    for channel in correct_channels:
        match = np.flatnonzero(np.core.defchararray.find(signal_labels, channel) != -1)
        if match.size > 0:
            match_indx.append(match)
    match_indx = np.unique(np.concatenate(match_indx))

    n_valid_signals = match_indx.shape[0]

    # Write edf file
    writer = EdfWriter(save_path, n_channels=n_valid_signals,
                       file_type=pyedflib.FILETYPE_EDFPLUS)

    channel_info = []
    data_list = []

    for i in np.arange(n):

        if i not in match_indx:
            continue

            # if not valid_chs[i]:
            # continue

        ch_dict = {'label': signal_labels[i],
                   'dimension': f.getPhysicalDimension(i),
                   'sample_rate': f.getSampleFrequency(i),
                   'physical_max': f.getPhysicalMaximum(i),
                   'physical_min': f.getPhysicalMinimum(i),
                   'digital_max': 32767.,
                   'digital_min': -32768.,
                   'transducer': f.getTransducer(i),
                   'prefilter': f.getPrefilter(i)}

        channel_info.append(ch_dict)
        data_list.append(f.readSignal(i))

    # Copy old file's annotation
    for annotation in f.read_annotation():
        writer.writeAnnotation(annotation[0], -1, annotation[2])
    writer.setHeader(f.getHeader())

    # Remove line noise notch filter
    sfreq = f.samplefrequency(0)
    time = np.arange(0, data_list[0].shape[0]) / sfreq
    coords = {'channels': signal_labels[match_indx],
              'time': time,
              'samplerate': sfreq}
    ts = TimeSeries.create(data_list, sfreq, coords=coords, dims=['channels', 'time'])
    bf = ButterworthFilter(timeseries=ts, freq_range=[58, 62], order=4, filt_type='stop')
    line_noise_filtered = bf.filter()

    # Create avg reference
    ch_mean = line_noise_filtered.mean('channels')
    avg_ref = line_noise_filtered - ch_mean

    # Save notched global avg as output
    data = [x for x in avg_ref.data]
    assert (len(data) == n_valid_signals)
    writer.setSignalHeaders(channel_info)
    writer.writeSamples(data)

    # Close writer/reader
    writer.close()
    f._close()


if __name__ == '__main__':
    from time import time
    from cmlreaders import CMLReader

    reader = CMLReader(subject='R1207J', experiment="FR1", session=1,
                       localization=0, montage=0)

    df = reader.load('contacts')
    ch_arr = np.array(df['label'])
    path = '/data/eeg/LFSleep/RawData/R1207J/8_17_16/8_17_16_0104_0234.edf'
    save = 'test.edf'
    start = time()
    preprocess_raw_edf(file_path=path, correct_channels=ch_arr, save_path=save)
    print('Total time: ', time() - start)


import numpy as np

from visbrain.gui import Brain
from visbrain.objects import TimeSeries3DObj, SourceObj
from visbrain.io import download_file
from Clumsy import get_sub_tal, TimeSeriesLF, rolling_window
__all__ = []
original = False
# Load the xyz coordinates and corresponding subject name :
if original:
    s_xyz = np.load(download_file('xyz_sample.npz', astype='example_data'))['xyz']
    s_xyz = s_xyz[4::25, ...]
    s_text = [str(k) for k in range(s_xyz.shape[0])]
    s_textsize = 1.5
else:
    #s_xyz = np.load(download_file('xyz_sample.npz', astype='example_data'))['xyz']
    mp,bp,tal = get_sub_tal('R1207J', 'FR1', True)
    x = tal['atlases']['ind']['x']
    y = tal['atlases']['ind']['y']
    z = tal['atlases']['ind']['z']
    arr = np.vstack((x,y,z))

    choice = np.random.choice(np.arange(tal.shape[0]), 25)
    s_xyz = arr[choice]
    #s_xyz = s_xyz[4::25, ...]
    s_text = [str(k) for k in choice]
    s_textsize = 1.5

"""Define a source object
"""
s_obj = SourceObj('MySources', s_xyz, symbol='disc', color='green')

"""Define the time-series data
"""
sf = 100.#512.                   # Sampling frequency
n_time_points = 700         # Number of time points
n_sources = s_xyz.shape[0]  # Number of sources
time = np.mgrid[0:n_sources, 0:n_time_points][1] / sf  # Time vector
# Randomize the amplitude and the phase of sine :
amp = np.random.randint(2, 20, n_sources).reshape(-1, 1)
pha = np.random.randint(1, 7, n_sources).reshape(-1, 1)
# Build the time series of shape (n_sources, n_time_points) :
ts_data = amp * np.sin(2 * np.pi * pha * time)
ts_data += np.random.randn(n_sources, n_time_points)

# Use a boolean vector to hide/display time-series :
ts_to_mask = [5, 7, 11, 3, 14, 17, 22, 23]
ts_select = np.ones((s_xyz.shape[0],), dtype=bool)
ts_select[ts_to_mask] = False

# Time-series (TS) graphical properties :
ts_amp = 5.4            # TS graphical amplitude
ts_width = 15.7         # TS graphical width
ts_color = 'orange'     # TS color
ts_dxyz = (1., 2., 5.)  # TS offset along the (x, y, z) axes
ts_lw = 2.2             # TS line-width

"""Define the 3-D time-series object
"""
ts = TimeSeries3DObj('Ts1', ts_data, s_xyz, select=ts_select, ts_amp=ts_amp,
                     ts_width=ts_width, line_width=ts_lw, translate=ts_dxyz,
                     color=ts_color)

vb = Brain(time_series_obj=ts, source_obj=s_obj)
vb.show()


from cmlreaders import CMLReader
import os
import datetime
import numpy as np
import pandas as pd
__all__ = ['get_sleep_behavior_dataframe',
           'get_data_roots',
           'get_subj_exp_sess_date_matcher',
           'check_int_ram_formatting',
           'subject_from_path',
           'get_night_dates_before_and_after',
           'get_sleep_path',
           ]

# Describes the current /data/eeg setup for all subjects
def get_data_roots():
    """gets data root structure

    Returns
    -------

    """
    DATA_ROOTS = {
        'LFSleep': {

            'ProcData': {

            }
            , 'RawData': {

            }
        }
    }

    return DATA_ROOTS

DATA_ROOTS = {
    'LFSleep' : {

        'ProcData': {

        }
        , 'RawData': {

        }
    }
}
def get_subj_exp_sess_date_matcher():
    """Describe relationships between experiments and when they took place for _sleep subjects

    Returns
    -------

    """
    SUBJ_EXP_SESS_DATE_MATCHER = {

        'R1207J': {

            'catfr1': {
                '0': '08_17_16_1543_1610',
                '1': '8_18_16_1530_1610'
            },

            'fr1': {
                '0': '8_20_16_1021_1105',
            }
        },

        'R1230J': {

            'fr1': {
                '0': '10_09_16_1400_1500',
                '2': '10_12_16_1138_1227',
                '4': '10_13_16_1520_1615',
                '5': '10_14_16_1017_1114',
                '6': '10_17_16_1320_1409',
            },

            'catfr1': {
                '0': '10_10_16_1220_1307',
            }
        },

        'R1293P': {

            'catfr1': {
                '0': '04_17_17_0955_1055',
            },

            'fr1': {
                '1': '04_18_17_1625_1715',
            },
        },

        'R1348J': {

            'catfr1': {
                '0': '09_26_17_1540_1630'
            },

            'dboy3': {
                '0': '09_23_17_1506',
                '1': '09_26_17_1449',
                '2': '09_27_17_1533',
            },

        },

        'R1384J': {

            'catfr1': {
                '0': '09_26_17_1540_1630'
            },

            'dboy3': {  # HOW SHOULD WE HANDLE??
                '0': '02_01_18_1825',
                '1': '02_03_18_1810',
                '2': '02_03_18_1812',
                '3': '02_07_18_1732',
                '4': '02_07_18_1734',
                '5': '02_10_18_1827',
                '6': '02_10_18_1828',
            },

            'dboy1': {
                '1': '02_17_18',
            },
        },
    }
    return SUBJ_EXP_SESS_DATE_MATCHER

SUBJ_EXP_SESS_DATE_MATCHER = {

    'R1207J': {

        'catfr1': {
            '0':'08_17_16_1543_1610',
            '1':'8_18_16_1530_1610'
        },

        'fr1':{
            '0':'8_20_16_1021_1105',
        }
    },

    'R1230J': {

        'fr1': {
            '0':'10_09_16_1400_1500',
            '2':'10_12_16_1138_1227',
            '4':'10_13_16_1520_1615',
            '5':'10_14_16_1017_1114',
            '6':'10_17_16_1320_1409',
        },

        'catfr1': {
            '0': '10_10_16_1220_1307',
        }
    },

    'R1293P': {

        'catfr1': {
            '0': '04_17_17_0955_1055',
        },

        'fr1' : {
            '1' : '04_18_17_1625_1715',
        },
    },

    'R1348J' : {

        'catfr1': {
          '0' : '09_26_17_1540_1630'
        },

        'dboy3': {
            '0': '09_23_17_1506',
            '1': '09_26_17_1449',
            '2': '09_27_17_1533',
        },

    },

    'R1384J' : {

        'catfr1': {
          '0' : '09_26_17_1540_1630'
        },

        'dboy3': { # HOW SHOULD WE HANDLE??
            '0':'02_01_18_1825',
            '1':'02_03_18_1810',
            '2':'02_03_18_1812',
            '3':'02_07_18_1732',
            '4':'02_07_18_1734',
            '5':'02_10_18_1827',
            '6':'02_10_18_1828',
        },

        'dboy1' : {
            '1': '02_17_18',
        },
    },
}
# SHOULD ALL OF THE ABOVE BE IN A DIFFERENT FILE?
def check_int_ram_formatting(s):
    """Check to see if inputted string matches RAM format

    Utility function used in subject_from_path
    Parameters
    ----------
    s: str, various dir names

    Returns
    -------

    """
    if len(s) == 0: # Nothing passed from split?
        return False
    # Check if it matches RAM formatting
    if s[0] in ('R') and s[-1] in ('J', 'P'): # Jefferson, Penn. E.g. R1207J
        return s[1:-1].isdigit()
    return s.isdigit()

def subject_from_path(path):
    """Given a path will extract out the subject ID if it is formatted as RAM structure

    Parameters
    ----------
    path: str, path of relevant data structure containing a subject folder

    Returns
    -------
    subject ID: str
    """
    for split in path.split('/'):
        valid = check_int_ram_formatting(split)
        if valid:
            return split


def get_night_dates_before_and_after(subject, experiment = None, session = None):
    """Returns the relevant dates for a subject given either their experiment and session or the date and time

    Parameters
    ----------
    subject: str,
             subject ID, e.g. 'R1111M'
    experiment: str, by default None
            experiment tag, valid: 'FR1', 'catFR1', dboy1, dboy3
    session: int, by default None
            session number for the experiment

    Returns
    -------
    before: datetime.datetime object of the day before
    after: datetime.datetime object of the day after
    """
    global SUBJ_EXP_SESS_DATE_MATCHER
    s = SUBJ_EXP_SESS_DATE_MATCHER[subject][experiment][session]
    a = s.split('_')
    # Convert the literal path into int and check formatting
    month = int(a[0])
    day = int(a[1])
    year = int(a[2])
    if year < 2000: # E.g. 8_12_18 -> 8_12_2018
        year += 2000
    date = datetime.datetime(year=year, month=month, day=day)
    before = date - datetime.timedelta(days=1)
    after = date + datetime.timedelta(days=1)
    return before, after


def get_sleep_path(subject, experiment = None, session = None,  datetimes = None, resampled=100):
    """Returns the relevant _sleep path for a subject given either their experiment and session or the date and time

    Parameters
    ----------
    subject: str,
             subject ID, e.g. 'R1111M'
    experiment: str, by default None
            experiment tag, valid: 'FR1', 'catFR1', dBoy
    session: int, by default None
            session number for the experiment
    datetimes: list, by default None
               two datetime.datetime objects consisting of the start and end point of the desired time
    resampled: int, by default 100 Hz
               for the given path collected, if there's a resampled version, then return that
               valid: 100, None
    Returns
    -------

    """
    # Check if we're using this locally or through rhino directly
    cwd = os.getcwd()
    local = '' if (cwd.split('/')[1][:4] == 'home') else '/Volumes/rhino'
    protocol = local + '/protocols/r1.json'
    global SUBJ_EXP_SESS_DATE_MATCHER
    return local + SUBJ_EXP_SESS_DATE_MATCHER[subject][experiment][session]


def _get_sleep_path(subject, experiment=None, session=None, datetimes=None):
    """Returns the relevant _sleep path for a subject given either their experiment and session or the date and time

    Parameters
    ----------
    subject: str,
             subject ID, e.g. 'R1111M'
    experiment: str, by default None
            experiment tag, valid: 'FR1', 'catFR1', dBoy
    session: int, by default None
            session number for the experiment
    datetimes: list, by default None
               two datetime.datetime objects consisting of the start and end point of the desired time

    Returns
    -------

    """
    # Check if we're using this locally or through rhino directly
    # cwd = os.getcwd()
    # local = '' if (cwd.split('/')[1][:4] == 'home') else '/Volumes/rhino'
    # protocol = local + '/protocols/r1.json'
    global SUBJ_EXP_SESS_DATE_MATCHER
    return SUBJ_EXP_SESS_DATE_MATCHER[subject][experiment][session]

#subjects = ["R1207J", "R1230J", 'R1293P', "R1384J", "R1348J", "R1398J"]

def get_sleep_behavior_dataframe(subjects):
    """Return the relevant dates, experiments, session and performance for inputted list of subjects
    subjects: np.array like, list of subjects to analyze
    """
    subjs_df = CMLReader.get_data_index()
    subj_d = {subject : subjs_df[subjs_df['subject']==subject][['experiment', 'session']] for subject in subjects}
    for k,v in enumerate(subj_d):
        subj_d[v] = subj_d[v][(subj_d[v]['experiment']=='FR1' )| (subj_d[v]['experiment']=='catFR1')]

    for k,v in enumerate(subj_d):
        performances = []
        dates = []
        for exp, sess in  np.array(subj_d[v]):
            rdr = CMLReader(subject=v, experiment=exp, session=sess)
            events = rdr.load('events')
            misses, hits = events[events['type']=='WORD']['recalled'].value_counts()
            performance = hits/(misses+hits)*100
            performances.append(performance)
            _,_,_,date,_ = events['eegfile'].iloc[11].split('_')
            dates.append(date)
        subj_d[v]['subject'] = v
        subj_d[v]['performance'] = performances
        subj_d[v]['date'] = dates

    subj_df = pd.concat(subj_d.values())



import numpy as np
from ptsa.data.common import get_axis_index
from ptsa.data.timeseries import TimeSeries
from scipy import signal

from Clumsy.signal.timeseriesLF import TimeSeriesLF

__all__ = ['psd']

def psd(x, Fs=None, method='mean', window='hann', nperseg=None, noverlap=None, filtlen=1.):
    """

    Estimating the power spectral density (PSD) of a time series from short-time Fourier
    Transform (mean, median), or the entire signal's FFT smoothed (medfilt).
    Parameters
    -----------
    x : array_like 1d
        Time series of measurement values.
    Fs : float, Hz
        Sampling frequency of the x time series.
    method : { 'mean', 'median', 'medfilt'}, optional
        Methods to calculate the PSD. Defaults to 'mean'.
            'mean' is the same as Welch's method (mean of STFT).
            'median' uses median of STFT instead of mean to minimize outlier effect.
            'medfilt' filters the entire signal raw FFT with a median filter to smooth.
    The next 3 parameters are only relevant for method = {'mean', 'median'}
    window : str or tuple or array_like, optional
        Desired window to use. Defaults to a Hann window.
            See scipy.signal.get_window for a list of windows and required parameters.
            If window is array_like, this array will be used as the window and its length must be nperseg.
    nperseg : int, optional
        Length of each segment, in number of samples. Defaults to None.
            If None, and window is str or tuple, is set to 1 second of data.
            If None, and window is array_like, is set to the length of the window.
    noverlap : int, optional
        Number of points to overlap between segments. If None, noverlap = nperseg // 2. Defaults to None.
    filtlen : float, Hz, optional
        (For medfilt method) Length of median filter in Hz.
    Returns
    -------
    freq : ndarray
        Array of sample frequencies.
    Pxx : ndarray
        Power spectral density of x.

    References
    ----------
    Mostly relies on scipy.signal.spectrogram and numpy.fft
    https://docs.scipy.org/doc/scipy/reference/generated/scipy.signal.spectrogram.html

    This is just a wrapper for timeseries object's around Voytek lab's neurodsp See:
    https://github.com/voytekresearch/neurodsp
    """
    is_timeseries = True if issubclass(type(x), TimeSeries) else False
    if (Fs is None and is_timeseries):
        Fs = int(x.samplerate.data)

    if method in ('mean', 'median'):
        # welch-style spectrum (mean/median of STFT)
        if nperseg is None:
            if isinstance(window, str) or isinstance(window, tuple):
                # window is a string or tuple, defaults to 1 second of data
                nperseg = int(Fs)
            else:
                # window is an array, defaults to window length
                nperseg = len(window)
        else:
            nperseg = int(nperseg)

        if noverlap is not None:
            noverlap = int(noverlap)

        # call signal.spectrogram function in scipy to compute STFT
        freq, _, spg = signal.spectrogram(x, Fs, window, nperseg, noverlap)
        if method is 'mean':
            Pxx = np.mean(spg, axis=-1)
        elif method is 'median':
            Pxx = np.median(spg, axis=-1)

    elif method is 'medfilt':
        # median filtered FFT spectrum
        # take the positive half of the spectrum since it's symmetrical
        FT = np.fft.fft(x)[:int(np.ceil(len(x) / 2.))]
        freq = np.fft.fftfreq(
            len(x), 1. / Fs)[:int(np.ceil(len(x) / 2.))]  # get freq axis
        # convert median filter length from Hz to samples
        filtlen_samp = int(int(filtlen / (freq[1] - freq[0])) / 2 * 2 + 1)
        Pxx = signal.medfilt(np.abs(FT) ** 2. / (Fs * len(x)), filtlen_samp)

    else:
        raise ValueError('Unknown PSD method: %s' % method)

    if is_timeseries:
        shell = x.sel(time=(x['time'].data[:1]))
        coords = {k: v for k, v in list(shell.coords.items())}
        coords['frequency'] = freq
        dims = shell.dims + ('frequency',)
        data = np.expand_dims(Pxx, axis=get_axis_index(shell, 'time'))
        powers_ts = TimeSeriesLF(data=data, dims=dims, coords=coords)
        return powers_ts

    return freq, Pxx


from .base import *
from .data_organization import *
from .detect import *
from .detection_utils import *
from .slowwaves import *
from .staging import *
#from .timeseries_brain_plot import *
from .transform_signal import *
#from .visbrain_gui import *


from ptsa.data.timeseries import TimeSeries
from scipy.ndimage.filters import gaussian_filter1d

__all__ = ['gaussian_smooth']

def gaussian_smooth(data, sampling_frequency=None, sigma=0.004, axis=-1, truncate=8):
    '''1D convolution of the data with a Gaussian.

    The standard deviation of the gaussian is in the units of the sampling
    frequency. The function is just a wrapper around scipy's
    `gaussian_filter1d`, The support is truncated at 8 by default, instead
    of 4 in `gaussian_filter1d`

    Parameters
    ----------
    data : array_like
    sigma : float
    sampling_frequency : int
    axis : int, optional
    truncate : int, optional

    Returns
    -------
    smoothed_data : array_like

    '''

    if ((sampling_frequency is None) & (type(data) == TimeSeries)):
        sampling_frequency = float(data['samplerate'])

    filtered = gaussian_filter1d(data, sigma * sampling_frequency,
                                 truncate=truncate, axis=axis, mode='constant')

    if type(data) == TimeSeries:
        # Create a copy of timeseries
        copy = TimeSeries(data=filtered, dims=data.dims, coords=data.coords)
        return copy

    return filtered


import numpy as np
import math

__all__ = ['fill_between_stderr','find_nearest', 'round_down', 'list_of_dict_to_dict']


def fill_between_stderr(arr):
    """
    arr: array of all subjects, axis 0 = subjects
    arr_tstat: array of plotting falls we'll use for the fillbetween
    """
    mean = np.nanmean(arr,axis=0)
    std_err_mean = np.nanstd(arr,axis=0) / ( np.sqrt( len(arr) ) )
    y1, y2 = np.array( (mean+std_err_mean) ), np.array( (mean-std_err_mean) )
    return y1, y2

def find_nearest(array, value, return_index_not_value=True, is_sorted=True):
    """Given an array and a value, returns either the index or value of the nearest match

    Parameters
    ----------
    array: np.array, array of values to check for matches
    value: int/float, value to find the closest match to
    return_index_not_value: bool, whether to return the index(True) or the value (False)
        of the found match
    is_sorted: bool, whether the array is sorted in order of values
    Returns
    -------
    Either the index or value of the nearest match
    """
    if is_sorted:
        idx = np.searchsorted(array, value, side="left")

        if ((idx > 0) and (idx == len(array)) or (math.fabs(value - array[idx - 1]) < math.fabs(value - array[idx]))):

            if not return_index_not_value:
                return array[idx - 1]

            if return_index_not_value:
                return idx - 1

        else:

            if not return_index_not_value:
                return array[idx]

            if return_index_not_value:
                return idx

    elif not is_sorted:
        idx = (np.abs(array - value)).argmin()

        if not return_index_not_value:
            return array[idx]

        if return_index_not_value:
            return idx


def round_down(n):
    """"Rounds number n to nearest 100th place

    Parameters
    ----------
    n: int/float, a number to round down

    Returns
    -------
    n: float, n rounded to nearest 100th place
    """
    return float(int(n / 100) * 100)

def list_of_dict_to_dict(d_list):
    '''
    Take a list of dictionaries and turn it into one dictionary
    ------
    INPUTS:
    d_list: a list of dictionaries
    ------
    RETURNS:
    d: a dictionary that's a merge of the values of the dictionaries in d_list
    '''
    d = {}
    # Go through dicts, add values into d
    for dictionaries in d_list:
        for keys, values in enumerate(dictionaries):
            if values not in d:
                d[values] = []
            if values in d:
                d[values].append(dictionaries[values])
    return d


import numpy as np
from numpy import (sqrt, mean, square, floor, zeros)

__all__ = [
    'root_mean_square',
    'RMS_amplitude_detector',
    'RMS_amplitude_detector_fast',
    'valid_rms_index'
]
def rolling_window(arr, window):
    """Returns an array of windowed sections of passed array a

    Parameters
    ----------
    arr: array, array to get windows of
    window: int, length of window

    Returns
    -------
    windowed array

    See Also http://www.rigtorp.se/2011/01/01/rolling-statistics-numpy.html
    """
    shape = arr.shape[:-1] + (arr.shape[-1] - window + 1, window)
    strides = arr.strides + (arr.strides[-1],)
    return np.lib.stride_tricks.as_strided(arr, shape=shape, strides=strides)


def root_mean_square(arr, axis=-1):
    """Calculates a root mean square on inputted array

    Parameters
    ----------
    arr: np.array, array to get root mean square of
    axis: int, by default -1, axis to apply root mean square along

    Returns
    -------
    root mean square of the array
    """
    # from numpy import (mean, sqrt, square)
    return sqrt(mean(square(arr), axis))



def RMS_amplitude_detector_fast(arr, duration=.2, fs=100.):
    """Applies a root mean square sliding window along an inputted array

    Parameters
    ----------
    arr: np.array, array to get root mean square of
    duration: float, desired duration in seconds
    fs: float, sample rate of array

    Returns
    -------
    root mean square of array
    """
    # from numpy import floor
    win = int(floor(fs * duration))
    rms = root_mean_square(rolling_window(np.array(arr), window=win))
    return rms


def valid_rms_index(rms, percentile=95):
    indx_above_thres = np.where(rms >= np.percentile(rms, percentile))
    return indx_above_thres


# SO SLOW!!!!
def RMS_amplitude_detector(data, duration=.2, fs=100.):
    """Computes a moving root mean square"""
    # from numpy import (floor, mean, sqrt, square, zeros)
    halfdur = int(floor(fs * duration / 2))
    lendat = len(data)
    rms = zeros((lendat))
    arr = data.data  # numpy array faster than timeseries indexing
    for i in range(lendat):
        rms[i] = sqrt(mean(square(
            arr[max(0, i - halfdur):min(lendat, i + halfdur)])))
    return rms


from __future__ import print_function
from itertools import chain
from collections import deque
import pickle
from six.moves import cPickle as cPickle
import sys
try:
    from reprlib import repr
except ImportError:
    print('Could not find rprlib cannot import repr')
    pass

__all__ = ['Save_Data', 'Load_Data', 'get_total_size']

def Save_Data(Object, path):
    """Saves arbitrary objects using pickling to desired path

    Parameters
    ----------
    Object: object, an arbitrary thing to save
    path: str, file path to Object to
    """
    with open((path), 'wb') as handle:
        pickle.dump(Object, handle, protocol=pickle.HIGHEST_PROTOCOL)
    print('saved successfully, file save as:\n' + path)

def Load_Data(path):
    """Loads objects using pickling

    Parameters
    ----------
    path: str, file path to load from

    Returns
    -------
    loaded_data: object, returned object
    """
    with open(path, 'rb') as handle:
        loaded_data = pickle.load(handle)
        return loaded_data

def sizecheck(object):
    """Check size of object"""
    # Old, should be depreciated/removed
    mydict = {'object': object}
    mydict_as_string = cPickle.dumps(mydict)
    print(sys.getsizeof(mydict_as_string))


def get_total_size(o, handlers={}, verbose=False):
    """ Returns the approximate memory footprint an object and all of its contents.
    Automatically finds the contents of the following builtin containers and
    their subclasses:  tuple, list, deque, dict, set and frozenset.
    To search other containers, add handlers to iterate over their contents:
        handlers = {SomeContainerClass: iter,
                    OtherContainerClass: OtherContainerClass.get_elements}


    Parameters
    ----------
    o: Object
    handlers:
    verbose: bool, whether to print out statements.

    Returns
    -------
    size


    ##### Example call #####

    #if __name__ == '__main__':
        #d = dict(a=power_ts)  # a=1, b=2, c=3, d=[4,5,6,7], e='a string of chars')
        #x = (total_size(d, verbose=True))
        #print(x)  # *1e-9)

    """

    dict_handler = lambda d: chain.from_iterable(d.items())
    all_handlers = {tuple: iter,
                    list: iter,
                    deque: iter,
                    dict: dict_handler,
                    set: iter,
                    frozenset: iter,
                    }
    all_handlers.update(handlers)  # User handlers take precedence
    seen = set()  # Track which object id's have already been seen
    default_size = sys.getsizeof(0)  # Estimate sizeof object without __sizeof__

    def sizeof(o):
        if id(o) in seen:  # do not double count the same object
            return 0
        seen.add(id(o))
        s = sys.getsizeof(o, default_size)

        if verbose:
            print(s, type(o), repr(o), file=sys.stderr)

        for typ, handler in all_handlers.items():
            if isinstance(o, typ):
                s += sum(map(sizeof, handler(o)))
                break
        return s

    return sizeof(o)


"""base.py Script for base detector that other detectors inherit from"""
from ptsa.data.timeseries import TimeSeries
from ptsa.data.common import get_axis_index
from scipy.stats import zscore
import numpy as np
import pandas as pd
from Clumsy import rolling_window_full, ButterworthFilter
from numba import jit
import traits.api
from .detection_utils import jit_find_containing_intervals, find_consecutive_data

__all__ = [
    #'jit_find_containing_intervals',
    'BaseDetector',
    #'find_consecutive_data'
]
# --------------> Base Detection Object
class BaseDetector(traits.api.HasTraits):
    """

    """

    valid = {'spindles': ['rms'],
             'ripples': ['buzsaki'],
             'slow-waves': ['amplitude', 'duration', 'fast'],
             'ied': ['buzsaki']
             }

    time_series = traits.api.Instance(TimeSeries)
    event_type = traits.api.Str
    method = traits.api.Str

    def __init__(self, time_series, event_type=None, method=None):
        """

        Parameters
        ----------
        time_series:TimeSeriesLF/ TimeSeries object,
        event_type
        method
        """
        if not issubclass(type(time_series), TimeSeries):
            raise(TypeError('Please input a time_series object'))

        super(BaseDetector, self).__init__()
        self.time_series = time_series

        #if not event_type in BaseDetector.valid:
            #raise(ValueError('inputted event_type {} not in {}'.format(event_type, BaseDetector.valid)))
        self.event_type = event_type

        #if not method in BaseDetector.valid[event_type]:
            #raise (ValueError('inputted method {} not in {}'.format(method, BaseDetector.valid[event_type])))
        self.method = method
        return

    ###### General Utility Functions #####
    @staticmethod
    def remove_linenoise(ts, harmonic=1, line_noise=60):
        for i in np.arange(1, harmonic+1):
            freqs = ((line_noise*i)-2, (line_noise*i)+2)
            ts = ts.filter_with(ButterworthFilter, freq_range=freqs, order=4, filt_type='stop', filter_output='sos')
        return ts

    def _dim_to_axis(self, dim):
        return get_axis_index(self.time_series, dim)

    def _duration_to_samples(self, duration):
        return self.time_series._TimeSeries__duration_to_samples(duration)

    @staticmethod
    def find_intervals(boolean_arr, min_sample, max_sample, contacts_df=None):
        """Find intervals where there is an amplitude and duration crossing

        Parameters
        ----------
        boolean_arr: np.array dtype=bool, shape = n chs x samples
                     Array indicating True if valid amplitude threshold or False if not
        min_sample: minimum time in samples to count as a valid duration interval
        max_sample: int or None,
                    minimum time in samples to count as a valid duration interval
        contacts_df: CMLReader contacts data frame

        Returns
        -------

        """

        try:
            fs_roi = contacts_df['ind.region']
            stein_roi = contacts_df['stein.region']
            hemi = ['left' if x == -1 else 'right' for x in np.sign(contacts_df['ind.x'])]
            channels = contacts_df['label']

        except AttributeError:
            fs_roi = ''
            stein_roi = ''
            hemi = ''
            channels = ''

        except Exception as e: # Check if AttributeError resolves everything
            print(e)
            fs_roi = ''
            stein_roi = ''
            hemi = ''
            channels = ''

        dataframe = []
        for index, ch_arr in enumerate(boolean_arr):
            start, stops = jit_find_containing_intervals(ch_arr, min_sample)
            df = pd.DataFrame(start, columns=['start'])
            try:
                df['stop'] = stops
            except ValueError as e:
                if len(start) != len(stops):  # Should occur when none are found
                    continue
                print(e)
            df['duration'] = df['stop'] - df['start']
            df['channel'] = channels[index]
            df['fs region'] = fs_roi[index]
            df['hemisphere'] = hemi[index]
            df['stein region'] = stein_roi[index]
            if max_sample is not None:
                df = df[df['duration'] < max_sample]
            dataframe.append(df)
        try:
            return pd.concat(dataframe)
        except ValueError as e:
            print(e)
            return dataframe

    ###### Thresholding Functions #####

    def rootmeansquare(self, window, asteps=None):
        """Applies a moving root mean square of window seconds over timeseries

        Parameters
        ----------
        window: float/int, Time in seconds to apply window over
        asteps: float/int, Time in seconds to allow window non-overlap
                 Aligned at the last axis, new steps for the original array, ie. for
                 creation of non-overlapping windows. (Equivalent to slicing result)

        Returns
        -------
        rms: np.array, array of root mean square
        """
        window_size = self.time_series._TimeSeries__duration_to_samples(window)
        asteps = self.time_series._TimeSeries__duration_to_samples(asteps)
        # Rolling root mean square on the data
        # rolled_data = rolling_window(self.time_series.data ** 2, window_size)
        rolled_data = rolling_window_full(self.time_series.data ** 2, window=window_size, asteps=asteps)
        rms = np.mean(rolled_data, -1)
        rms = np.sqrt(rms)
        return rms

    @staticmethod
    def percentile_threshold(data, threshold=92, axis=1):
        """Apply a percentile based threshold to the data

        Parameters
        ----------
        data: array_like
              data to apply threshold on
        threshold: array_like of float
                   Percentile or sequence of percentiles to compute, which must be between
                   0 and 100 inclusive.
        axis: int
              axis to threshold over

        Returns
        -------
        bool_arr, locations where inputted data is above desired threshold
        """
        # Use percentile thresholding
        percentiles = np.percentile(data, q=threshold, axis=axis)
        return data > percentiles[:, None]

    @staticmethod
    def zscore_threshold(data, threshold=3, axis=-1):
        """Use z-score to find out where data is at n threshold above the standard deviation of the mean

        Parameters
        ----------
        data
        threshold
        axis

        Returns
        -------
        bool_arr, locations where inputted data is above desired threshold
        """
        zdata = zscore(data, axis=axis)
        #mean = np.mean(zdata, axis=axis)
        #std = np.std(zdata, axis=axis)
        #threshold = mean + (std * threshold)
        return zdata > threshold

    @staticmethod
    def threshold_value_from_array(data, threshold):
        return data > threshold




from .butterworth import *
#from .circlestats import *
from .datachop import *
from .fft import *
from .guassiansmoothing import *
from .hilbert import *
from .rolling import *
from .rootmeansquare import *
from .timeseriesLF import *



from glob import glob
from time import time
from Clumsy import TimeSeriesLF
from Clumsy import get_sub_tal
from visbrain import gui
import numpy as np
from cmlreaders import CMLReader

__all__ = []
start = time()

paths = glob('/Users/loganfickling/Clumsy/data/*.h5')
paths_done = [
    #'/Users/loganfickling/Clumsy/data/8_17_16_1604_1734_100hz.h5',


]

mp,bp,tal = get_sub_tal("R1207J",'FR1',True)
reader = CMLReader(subject='R1207J', experiment='FR1', session=1, rootdir='/Volumes/rhino')
pairs = reader.load('pairs')
contacts = reader.load('contacts')
electrode_categories = reader.load('electrode_categories')
remove = np.concatenate(list(electrode_categories.values()))
labels = pairs['label']
x = np.array(list(map(lambda s:s.split('-'), labels)))
ch0 = np.in1d(x[:,0], np.intersect1d(x[:,0], remove))
ch1 = np.in1d(x[:,1], np.intersect1d(x[:,1], remove))
healthy_bp = bp[(~ch0 & ~ch1)]


for path in paths:
    if path in paths_done:
        continue
    path = '/Users/loganfickling/Clumsy/data/8_17_16_1604_1734_100hz.h5'
    print('Starting path %s' % path)
    ts = TimeSeriesLF.from_hdf(path)
    data = np.array(ts.data)
    data *= 10**6 # Conversion due to units
    sf = 100 # Could also access in ts
    chan = np.arange(len(ts.channels.data)).astype(str)
    print('Time to start gui', time()-start)
    gui.Sleep(data=data, channels=chan, sf=sf, hypno=None).show()
    print('Close gui on path %s' % path)

    break



# Script to use cmlreaders to easily get data on subjects
from cmlreaders import CMLReader, get_data_index
from ptsa.data.readers import JsonIndexReader, TalReader
import os

__all__ = [
    'get_subjects',
    'get_sessions',
    'get_ram_experiments',
    'get_sub_tal'
]

def get_subjects(experiment):
    """Returns an array of all subjects who participated in the experiment

    Parameters
    ----------
    experiment: str, must be 'CatFR1', 'FR1', 'pyFR' or 'ltpFR2'

    Returns
    -------
    All valid subjects
    """
    if experiment.lower() in ('catfr1', 'fr1'):
        dataframe = get_data_index("r1")

    elif experiment.lower() in ('pyfr'):
        dataframe = get_data_index("pyfr")

    elif experiment.lower() in ('ltp', 'ltpfr2'):
        dataframe = get_data_index("ltp")
    else:

        try:
            dataframe = get_data_index(experiment)
        except ValueError as e:
            print(e)
            return

    return dataframe[dataframe['experiment'] == experiment]['subject'].unique()


def get_sessions(subject, experiment):
    """For a given subject and experiment return an array of all valid sessions

    Parameters
    ----------
    subject: str, must be a valid pyFR or FR1 or catFR1 subject
    experiment: str, must be a valid experiment.
                Currently implemented experiments:
                FR1, catFR1, pyFR

    Returns
    -------

    """
    RAM = ['catFR1', 'FR1']
    if experiment in RAM:
        dataframe = get_data_index("r1")
        return dataframe[(dataframe['experiment'] == experiment)
                         & (dataframe['subject'] == subject)]['session'].unique()
    elif experiment == 'pyFR':
        dataframe = get_data_index("pyfr")
        return dataframe[(dataframe['experiment'] == experiment)
                         & (dataframe['subject'] == subject)]['session'].unique()

def get_ram_experiments(subject):
    """Given a subject from RAM returns the experiments

    Parameters
    ----------
    subject: str, RAM subject id to get experiments for

    Returns
    -------
    all valid experiments for the subject
    """
    dataframe = get_data_index("r1")
    return dataframe[dataframe['subject']==subject]['experiment'].unique()


def get_sub_tal(subject, experiment, return_channels=False):
    """Returns a subject's talairach using TalReader and JsonIndexReader
    -----
    INPUTS:
    -----
    subject: str, subject ID, e.g. 'R1111M'
    experiment: str, experiment, e.g. 'FR1', 'catFR1'
    return_channels: bool, default = False, whether to return arrays of
                     monopolar and bipolar channels used for EEGReader
    ------
    OUTPUTS if return_channels is False:
    ------
    tal_reader.read(): np.recarray, an array containing relevant values
                       for electrode localization
    ------
    OUTPUTS if return_channels is True:
    ------
    mp: np.recarray, monopolar channels used for EEGReader
    bp: np.recarray, bipolar channels used for EEGReader
    tal_reader.read(): np.recarray, an array containing relevant values
                       for electrode localization
    """

    # Check if we're using this locally or through rhino directly
    cwd = os.getcwd()
    local = '' if (cwd.split('/')[1][:4] == 'home') else '/Volumes/rhino'
    protocol = local + '/protocols/r1.json'

    # Load the protocol
    jr = JsonIndexReader(protocol)

    # Get the path for the TalReader
    pairs_path = jr.get_value('pairs', subject=subject, experiment=experiment)

    # Create an instance of the TalReader object
    tal_reader = TalReader(filename=pairs_path)

    # If desired, also return the monopolar and bipolar channels
    if return_channels:
        mp = tal_reader.get_monopolar_channels()
        bp = tal_reader.get_bipolar_pairs()
        return mp, bp, tal_reader.read()

    # Otherwise just return the tal
    return tal_reader.read()

if __name__ == '__main__':
    mp, bp, bp_tal = get_sub_tal(subject='R1207J', experiment='FR1', return_channels=True)


try:
    from matplotlib.colors import LinearSegmentedColormap, ListedColormap
    from matplotlib import pyplot as plt
except ImportError as e:
    print('Encountering ImportErrors due to issues with python package matplotlib',
          'not playing nice with an OS X background',
          '\nAttempting to fix...')
    plt = None
    LinearSegmentedColormap = None
    ListedColormap = None
    print('Fixed')
import numpy as np

__all__ = ['wes_palettes',
           'shiftedColorMap',
           'create_colormap',
           'discrete_colormap',
           'rgb_to_hex',
           'hex_to_rgb',
           ]

wes_palettes = {
  'BottleRocket1' : ["#A42820", "#5F5647", "#9B110E", "#3F5151", "#4E2A1E", "#550307", "#0C1707"],
  'BottleRocket2' : ["#FAD510", "#CB2314", "#273046", "#354823", "#1E1E1E"],
  'Rushmore1' : ["#E1BD6D", "#EABE94", "#0B775E", "#35274A" ,"#F2300F"],
  'Rushmore' : ["#E1BD6D", "#EABE94", "#0B775E", "#35274A" ,"#F2300F"],
  'Royal1' : ["#899DA4", "#C93312", "#FAEFD1", "#DC863B"],
  'Royal2' : ["#9A8822", "#F5CDB4", "#F8AFA8", "#FDDDA0", "#74A089"],
  'Zissou1' : ["#3B9AB2", "#78B7C5", "#EBCC2A", "#E1AF00", "#F21A00"],
  'Darjeeling1' : ["#FF0000", "#00A08A", "#F2AD00", "#F98400", "#5BBCD6"],
  'Darjeeling2' : ["#ECCBAE", "#046C9A", "#D69C4E", "#ABDDDE", "#000000"],
  'Chevalier1' : ["#446455", "#FDD262", "#D3DDDC", "#C7B19C"],
  'FantasticFox1' : ["#DD8D29", "#E2D200", "#46ACC8", "#E58601", "#B40F20"],
  'Moonrise1' : ["#F3DF6C", "#CEAB07", "#D5D5D3", "#24281A"],
  'Moonrise2' : ["#798E87", "#C27D38", "#CCC591", "#29211F"],
  'Moonrise3' : ["#85D4E3", "#F4B5BD", "#9C964A", "#CDC08C", "#FAD77B"],
  'Cavalcanti1' : ["#D8B70A", "#02401B", "#A2A475", "#81A88D", "#972D15"],
  'GrandBudapest1' : ["#F1BB7B", "#FD6467", "#5B1A18", "#D67236"],
  'GrandBudapest2' : ["#E6A0C4", "#C6CDF7", "#D8A499", "#7294D4"],
  'IsleofDogs1' : ["#9986A5", "#79402E", "#CCBA72", "#0F0D0E", "#D9D0D3", "#8D8680"],
  'IsleofDogs2' : ["#EAD3BF", "#AA9486", "#B6854D", "#39312F", "#1C1718"]}

def shiftedColorMap(cmap, start=0, midpoint=0.5, stop=1.0, name='shiftedcmap'):
    '''
    Function to offset the "center" of a colormap. Useful for
    data with a negative min and positive max and you want the
    middle of the colormap's dynamic range to be at zero

    Input
    -----
      cmap : The matplotlib colormap to be altered
      start : Offset from lowest point in the colormap's range.
          Defaults to 0.0 (no lower ofset). Should be between
          0.0 and `midpoint`.
      midpoint : The new center of the colormap. Defaults to
          0.5 (no shift). Should be between 0.0 and 1.0. In
          general, this should be  1 - vmax/(vmax + abs(vmin))
          For example if your data range from -15.0 to +5.0 and
          you want the center of the colormap at 0.0, `midpoint`
          should be set to  1 - 5/(5 + 15)) or 0.75
      stop : Offset from highets point in the colormap's range.
          Defaults to 1.0 (no upper ofset). Should be between
          `midpoint` and 1.0.
    '''
    cdict = {
        'red': [],
        'green': [],
        'blue': [],
        'alpha': []
    }

    # regular index to compute the colors
    reg_index = np.linspace(start, stop, 257)

    # shifted index to match the data
    shift_index = np.hstack([
        np.linspace(0.0, midpoint, 128, endpoint=False),
        np.linspace(midpoint, 1.0, 129, endpoint=True)
    ])

    for ri, si in zip(reg_index, shift_index):
        r, g, b, a = cmap(ri)

        cdict['red'].append((si, r, r))
        cdict['green'].append((si, g, g))
        cdict['blue'].append((si, b, b))
        cdict['alpha'].append((si, a, a))

    newcmap = LinearSegmentedColormap(name, cdict)
    plt.register_cmap(cmap=newcmap)

    return newcmap


def create_colormap(colors=None, cmap_name=None, bins=10):
    """Given a list of colors, creates a colormap with desired spacing

    Parameters
    ----------
    colors: list
            a list of colors to generate a colormap over
    cmap_name: str, by default None and 'my_list' is used
               Name of map to register
    bins: int, by default 10
          Number of steps to use in total map

    Returns
    -------

    """
    if colors is None:
        # cmap_name = 'MoonriseKingdomCustom'
        # colors = ['#cb654f', '#d3b1a7', '#cfcb9c', '#8cbea3', '#dfba47',
        # '#fad77b', '#cdc08c','#9c964a', '#f4b5bd', '#86d4e4']
        colors = ["#FF0000", "#00A08A", "#F2AD00", "#F98400", "#5BBCD6",
                  "#E1BD6D", "#EABE94", "#0B775E", "#35274A", "#F2300F"]
        if cmap_name is None:
            cmap_name = 'Rushmore1'
    if cmap_name is None:
        cmap_name = 'my_list'

    colormap = LinearSegmentedColormap.from_list(cmap_name, colors, N=bins)
    return colormap

def rgb_to_hex(rgb):
    """Converts a rgb value to hex code"""
    return '#%02x%02x%02x' % rgb

def hex_to_rgb(value):
    """converts a hexcode to rgb value"""
    value = value.lstrip('#')
    lv = len(value)
    return tuple(int(value[i:i + lv // 3], 16) for i in range(0, lv, lv // 3))


def discrete_colormap(color_pool = None, N=8):
    """create a color map with N (N<15) discrete colors and register it"""
    # define individual colors as hex values
    if color_pool is None:

        color_pool = [ '#bd2309', '#bbb12d', '#1480fa', '#14fa2f', '#000000',
                      '#faf214', '#2edfea', '#ea2ec4', '#ea2e40', '#cdcdcd',
                      '#577a4d', '#2e46c0', '#f59422', '#219774', '#8086d9' ]

    color_map = ListedColormap(color_pool[0:N], 'indexed')
    return color_map


import numpy as np
from numba import jit
from ptsa.data.timeseries import TimeSeries
from scipy.stats import zscore
__all__ = [

    'find_consecutive_data',
    'find_containing_intervals',
    'chop_intervals',
    'init_channel_events',
    'jit_find_containing_intervals'
]

def find_consecutive_data(data, stepsize=1):
    """Splits Data into a list of arrays where there is more than a single step of 1

    Parameters
    ----------
    data: np.array of zeroes and ones, the data to split
    stepsize: int, by default 1

    Returns
    -------
    list of split data
    """
    return np.split(data, np.where(np.diff(data) != stepsize)[0] + 1)


# --------> FUNCTIONS TO FIND WHERE THERES A CROSSING!!!!!

def zscore_amp_threshold(data, minimum_amplitude=3):
    """Applies z-score threshold on data"""
    z_data = zscore(data, -1)
    above_thres = z_data > minimum_amplitude
    return above_thres


def sampling_freq_to_minimum_duration(ts, desired_length_in_ms=20):
    """Computes the minimum window length that corresponds with the desired length in ms"""
    sr = float(ts['samplerate'])
    window_length = int((float(desired_length_in_ms) / 1000) / (1 / sr))
    return window_length


# Use numba to vastly improve the speed of this (x10) through the use of compiling
@jit(nopython=True)
def jit_find_containing_intervals(channel, duration):
    """Find intervals where there is an amplitude and duration crossing

    NOTES: Uses numba to vastly improve the speed of this (x10)
           through the use of just-in-time (jit) compiling of the code
           Use the function find_containing_intervals which acts as a wrapper on this
           rather than this function itself
    ------
    INPUTS:
    ------
    channel: bool array, True if above amplitude Threshold else False

    ------
    OUTPUTS:
    ------
    (n x 2) array of start, stop indicies
    """
    count = 0
    starts, stops = [], []

    for i, j in enumerate(channel):
        # Crossed the duration threshold!
        if count == duration:
            starts.append(i - duration)

        if i == 0:
            continue

        if j:  # If the channel crosses amplitude threshold, increment counter
            count += 1
            continue

        if not j:  # Otherwise reset it and it's of a sufficient duration add a stop
            if count >= duration:
                stops.append(i)
            count = 0
            continue

    return starts, stops


def find_containing_intervals(ts):
    """Find intervals where there is an amplitude and duration crossing
    ------
    INPUTS:
    ------
    ts: TimeSeriesX object; instantaneous amplitude envelope

    ------
    OUTPUTS:
    ------
    data: list, [ch x array(n x 2)] of start, stop indicies
    """
    duration = sampling_freq_to_minimum_duration(ts)
    bool_arr = zscore_amp_threshold(ts)
    data = [np.vstack((jit_find_containing_intervals(channel, duration))).T for channel in bool_arr]
    return data


def chop_intervals(intervals, timeseries, ch_index, return_timeseries=False):
    """Chops intervals produced by find_containing_intervals and returns a chopped timeseries for the ch
    -----
    INPUTS:
    intervals: output of find_containing_intervals()
    timeseries: TimeSeriesX, must be amplitude envelope
    ch_index: int, index to channel
    return_timeseries: whether or not to return it as a TimeSeriesX object (True) or list (False)
    ------
    OUTPUTS:
    chopped_channel: TimeSeriesX, timeseries chopped between amplitude and duration threshold crossing
    """
    channel_interval = intervals[ch_index]
    chopped_channel = []
    for start, stop in zip(channel_interval[:, 0], channel_interval[:, 1]):
        chopped_channel.append(timeseries[ch_index, start:stop])
    # Make it a TimeSeriesX object again
    if return_timeseries:
        chopped_channel = TimeSeries.concat(chopped_channel, 'time')
    return chopped_channel


def init_channel_events(intervals, ch_index, path, subject='R1293P'):
    """Creates a behavioral record array indiciating the start/stops of potential ripples
    """
    fields = np.dtype([('subject', 'O'), ('path', 'O'), ('start', int), ('stop', int), ('duration', int)])
    events = np.zeros(len(intervals[ch_index]), dtype=fields)
    events['start'] = intervals[ch_index][:, 0]
    events['stop'] = intervals[ch_index][:, 1]
    events['subject'] = np.array([subject] * len(intervals[ch_index]))
    events['path'] = np.array([path] * len(intervals[ch_index]))
    events['duration'] = events['stop'] - events['start']
    events = events.view(np.recarray)
    return events



from mne.filter import notch_filter as _notch_filter
from ptsa.data.common import get_axis_index
from ptsa.data.timeseries import TimeSeries
from Clumsy.signal.rolling import reorder_dims

import numpy as np

def notch_filter(timeseries, europe=False):
    """Wrapper around mne's notch_filter function

    Parameters
    ----------
    timeseries
    europe

    Returns
    -------
    notch filtered data (no line noise)

    Notes
    -------
    mne implicitly assumes that time is the last dimension in the timeseries
    """

    reordered = False
    if ((len(timeseries.shape) > 1)
            & (get_axis_index(timeseries, 'time') + 1 != len(timeseries.shape))):
        reordered = True

        timeseries = reorder_dims(timeseries, -1, get_axis_index(timeseries, 'time'))

    freqs = np.arange(60, 241, 60) if not europe else np.arange(50, 201, 50)
    filtered = _notch_filter(x=timeseries.data.astype(np.float64), Fs=timeseries['samplerate'], freqs=freqs)

    return TimeSeries(filtered, coords=timeseries.coords, dims=timeseries.dims)

if __name__ == '__main__':
    # Generate quasi-periodic data with an alpha burst
    from matplotlib import pyplot as plt
    np.random.seed(0)
    t = np.arange(0, 2, .001)
    x = 30 * np.random.randn(len(t)) + 50 * np.sin(t * 2 * np.pi * 1.3) + 30 * np.sin(
        t * 2 * np.pi * 18 - .3)  # background activity
    x[500:1000] += 80 * np.sin(t[500:1000] * 2 * np.pi * 10)  # add in alpha burst
    Fs = 1000

    # Define the frequency band of interest
    alpha = (8, 12)

    # Add in line noise
    y = x + (80 * np.sin(t * 2 * np.pi * 60))  # add in alpha burst
    ts = TimeSeries(data=y, coords={'time': t, 'samplerate': Fs}, dims=['time'])
    filtered=notch_filter(timeseries=ts)
    #print(x)
    fig, ax = plt.subplots(ncols=3, figsize=(12,4))
    ax[0].plot(x)
    ax[1].plot(ts)
    ax[2].plot(filtered)
    plt.show()


"""circstats.py -- circular statistics module
This software is licensed under the terms of the MIT License as
follows:
Copyright (c) 2013 Jessica B. Hamrick, Peter W. Battaglia
Permission is hereby granted, free of charge, to any person obtaining
a copy of this software and associated documentation files (the
"Software"), to deal in the Software without restriction, including
without limitation the rights to use, copy, modify, merge, publish,
distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to
the following conditions:
The above copyright notice and this permission notice shall be
included in all copies or substantial portions of the Software.
THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""

__all__ = [
    'difference',
    'wrapdiff',
    'normalize',
    'csr',
    'to_complex',
    'to_real',
    'mean',
    'nanmean',
    'resvec',
    'nanresvec',
    'var',
    'nanvar',
    'std',
    'nanstd',
    'confmean',
    'nanconfmean',
    'corrcc',
    'nancorrcc',
    'vmpar',
    'nanvmpar',
    'vmlogpdf',
    'vmpdf',
    'kappa',
    'nankappa'
]

import numpy as np
from scipy.stats import chi2
from scipy.special import iv


def difference(a1, a2, deg=False):
    """Compute the smallest difference between two angle arrays.
    Parameters
    ----------
    a1, a2 : np.ndarray
        The angle arrays to subtract
    deg : bool (default=False)
        Whether to compute the difference in degrees or radians
    Returns
    -------
    out : np.ndarray
        The difference between a1 and a2
    """

    diff = a1 - a2
    return wrapdiff(diff, deg=deg)


def wrapdiff(diff, deg=False):
    """Given an array of angle differences, make sure that they lie
    between -pi and pi.
    Parameters
    ----------
    diff : np.ndarray
        The angle difference array
    deg : bool (default=False)
        Whether the angles are in degrees or radians
    Returns
    -------
    out : np.ndarray
        The updated angle differences
    """

    if deg:
        base = 360
    else:
        base = np.pi * 2

    i = np.abs(diff) > (base / 2.0)
    out = diff.copy()
    out[i] -= np.sign(diff[i]) * base
    return out


def normalize(data, deg=False):
    """Make sure that all values in `data` lie between 0 and 2*pi (or
    0 and 360, for degrees).
    Parameters
    ----------
    data : number or np.array
        The value(s) to normalize
    deg : bool (default=False)
        Whether the values are in degrees or radians
    Returns
    -------
    out : number np.ndarray
        The normalized value(s)
    """

    if deg:
        base = 360
    else:
        base = np.pi * 2

    data = data % base
    return data


def csr(data, axis=None, nanrobust=False, deg=False):
    """Finds the mean cosines, sines, and radii of an array of angles
    along the given axis.
    Parameters
    ----------
    data : np.ndarray
        The angle array
    axis : int (default=None)
        The axis along which to take the mean computations
    nanrobust : bool (default=False)
        Ignore nans
    deg : bool (default=False)
        Whether the values are in degrees or radians
    Returns
    -------
    out : tuple
        (mean cosine, mean sine, mean radius)
    """

    if deg:
        data = np.radians(data)

    if nanrobust:
        n = np.sum(~np.isnan(data), axis=axis)
        C = np.nansum(np.cos(data), axis=axis) / n
        S = np.nansum(np.sin(data), axis=axis) / n
    else:
        C = np.mean(np.cos(data), axis=axis)
        S = np.mean(np.sin(data), axis=axis)

    R = (C ** 2 + S ** 2) ** 0.5

    return (C, S, R)


def to_complex(data, deg=False):
    """Convert an array of angles to complex values, e ** (data * 1j).
    Parameters
    ----------
    data : np.ndarray
        The angle values
    deg : bool (default=False)
        Whether the values are in degrees or radians
    Returns
    -------
    out : np.ndarray
        An array of complex-valued angles
    """

    if deg:
        d = np.radians(data)
    else:
        d = data.copy()
    return np.exp(d * 1j)


def to_real(data, deg=False):
    """Convert an array of complex values to angles.
    Parameters
    ----------
    data : np.ndarray
        The complex values
    deg : bool (default=False)
        Whether the values are in degrees or radians
    Returns
    -------
    out : np.ndarray
        An array of angles
    """

    return np.angle(data, deg=deg)


def _mean(alpha, nanrobust, axis=None, w=None, fconf=False):
    if w is None:
        w = np.ones(alpha.shape)

    # weight sum of cos & sin angles
    t = w * np.exp(1j * alpha)
    if nanrobust:
        r = np.nansum(t, axis=axis)
        w[np.isnan(t)] = 0
    else:
        r = np.sum(t, axis=axis)

    # obtain mean by
    #mu = normalize(np.angle(r))
    mu = np.angle(r)

    if fconf:
        # confidence limits if desired
        if nanrobust:
            t = nanconfmean(alpha, w=w, axis=axis)
        else:
            t = confmean(alpha, w=w, axis=axis)
        ul = mu + t
        ll = mu - t

        return (mu, ul, ll)
    else:
        return mu


def mean(alpha, axis=None, w=None, fconf=False, nancheck=True):
    """Calculate the mean of an array of angles using circular
    statistics.
    Parameters
    ----------
    alpha : np.ndarray
        The array of angles
    axis : int (default=None)
        The axis along which to take the mean
    w : np.ndarray (default=None)
        An array of weights, if a weighted mean is desired
    fconf : bool (default=False)
        Whether or not to return confidence intervals along with the
        means
    nancheck : bool (default=True)
        Whether or not to check for NaNs in the array
    Returns
    -------
    out : np.ndarray OR (np.ndarray, np.ndarray, np.ndarray)
        The resulting numpy array, or a tuple of the array, the upper
        confidence limit, and the lower confidence limit
    References
    ----------
    Berens, P. (2009). CircStat: A MATLAB Toolbox for Circular
      Statistics. Journal of Statistical Software, 31(10). Available
      from http://www.jstatsoft.org/v31/i10
    """

    if nancheck and np.isnan(alpha).any():
        raise(ValueError,
              "NaN(s) detected in the input array.  "
              "If you want to ignore this message, "
              "please set nancheck to False.  If you "
              "want to make the mean robust to NaNs, "
              "please use the nanmean function instead.")

    return _mean(alpha, False, axis=axis, w=w, fconf=fconf)


def nanmean(alpha, axis=None, w=None, fconf=False):
    """Calculate the mean of an array of angles using circular
    statistics.
    Parameters
    ----------
    alpha : np.ndarray
        The array of angles
    axis : int (default=None)
        The axis along which to take the mean
    w : np.ndarray (default=None)
        An array of weights, if a weighted mean is desired
    fconf : bool (default=False)
        Whether or not to return confidence intervals along with the
        means
    Returns
    -------
    out : np.ndarray OR (np.ndarray, np.ndarray, np.ndarray)
        The resulting numpy array, or a tuple of the array, the upper
        confidence limit, and the lower confidence limit
    References
    ----------
    Berens, P. (2009). CircStat: A MATLAB Toolbox for Circular
      Statistics. Journal of Statistical Software, 31(10). Available
      from http://www.jstatsoft.org/v31/i10
    """

    return _mean(alpha, True, axis=axis, w=w, fconf=fconf)


def _resvec(alpha, nanrobust, axis=None, w=None, d=0):
    if w is None:
        w = np.ones(alpha.shape)

    # weight sum of cos & sin angles
    t = w * np.exp(1j * alpha)
    if nanrobust:
        r = np.nansum(t, axis=axis)
        w[np.isnan(t)] = 0
    else:
        r = np.sum(t, axis=axis)

    # obtain length
    r = np.abs(r) / np.sum(w, axis=axis)

    # for data with known spacing, apply correction factor to correct
    # for bias in the estimation of r (see Zar, p. 601, equ. 26.16)
    if d != 0:
        r *= d / 2 / np.sin(d / 2)

    return r


def resvec(alpha, axis=None, w=None, d=0, nancheck=True):
    """Computes mean resultant vector length for circular data.
    Parameters
    ----------
    alpha : np.ndarray
        The array of angles
    axis : int (default=None)
        The axis along which to take the mean
    w : np.ndarray (default=None)
        An array of weights, if a weighted mean is desired
    d : number (default=0)
        The bin spacing, if the data is binned
    nancheck : bool (default=True)
        Whether or not to check for NaNs in the array
    Returns
    -------
    out : np.ndarray
        The resulting numpy array of resultant vector lengths
    References
    ----------
    Berens, P. (2009). CircStat: A MATLAB Toolbox for Circular
      Statistics. Journal of Statistical Software, 31(10). Available
      from http://www.jstatsoft.org/v31/i10
    """

    if nancheck and np.isnan(alpha).any():
        raise(ValueError,
              "NaN(s) detected in the input array.  If you "
              "want to ignore this message, please set nancheck "
              "to False.  If you want to make the mean robust to "
              "NaNs, please use the nanresvec function instead.")

    return _resvec(alpha, False, axis=axis, w=w, d=d)


def nanresvec(alpha, axis=None, w=None, d=0):
    """Computes mean resultant vector length for circular data.
    Parameters
    ----------
    alpha : np.ndarray
        The array of angles
    axis : int (default=None)
        The axis along which to take the mean
    w : np.ndarray (default=None)
        An array of weights, if a weighted mean is desired
    d : number (default=0)
        The bin spacing, if the data is binned
    Returns
    -------
    out : np.ndarray
        The resulting numpy array of resultant vector lengths
    References
    ----------
    Berens, P. (2009). CircStat: A MATLAB Toolbox for Circular
      Statistics. Journal of Statistical Software, 31(10). Available
      from http://www.jstatsoft.org/v31/i10
    """

    return _resvec(alpha, True, axis=axis, w=w, d=d)


def _var(alpha, nanrobust, axis=None, w=None, d=0, angvar=False):
    if w is None:
        w = np.ones(alpha.shape)

    # compute mean resultant vector length
    if nanrobust:
        r = nanresvec(alpha, axis=axis, w=w, d=d)
    else:
        r = resvec(alpha, axis=axis, w=w, d=d)

    # apply transformation to var
    S = 1 - r  # circular variance
    if angvar:
        var = 2 * S  # angular variance
    else:
        var = S

    return var


def var(alpha, axis=None, w=None, d=0, angvar=False, nancheck=True):
    """Computes circular variance for circular data.
    Parameters
    ----------
    alpha : np.ndarray
        The array of angles
    axis : int (default=None)
        The axis along which to take the mean
    w : np.ndarray (default=None)
        An array of weights, if a weighted mean is desired
    d : number (default=0)
        The bin spacing, if the data is binned
    angvar : bool (default=False):
        If True, calculate angular variance, otherwise calculate
        circular variance
    nancheck : bool (default=True)
        Whether or not to check for NaNs in the array
    Returns
    -------
    out : tuple (np.ndarray, np.ndarray)
        The calculated circular variance and angular variance
    References
    ----------
    Berens, P. (2009). CircStat: A MATLAB Toolbox for Circular
      Statistics. Journal of Statistical Software, 31(10). Available
      from http://www.jstatsoft.org/v31/i10
    """

    if nancheck and np.isnan(alpha).any():
        raise(ValueError,
              "NaN(s) detected in the input array.  If you want "
              "to ignore this message, please set nancheck to "
              "False.  If you want to make the mean robust to NaNs, "
              "please use the nanvar function instead.")

    return _var(alpha, False, axis=axis, w=w, d=d, angvar=angvar)


def nanvar(alpha, axis=None, w=None, d=0, angvar=False):
    """Computes circular variance for circular data.
    Parameters
    ----------
    alpha : np.ndarray
        The array of angles
    axis : int (default=None)
        The axis along which to take the mean
    w : np.ndarray (default=None)
        An array of weights, if a weighted mean is desired
    d : number (default=0)
        The bin spacing, if the data is binned
    angvar : bool (default=False):
        If True, calculate angular variance, otherwise calculate
        circular variance
    Returns
    -------
    out : tuple (np.ndarray, np.ndarray)
        The calculated circular variance and angular variance
    References
    ----------
    Berens, P. (2009). CircStat: A MATLAB Toolbox for Circular
      Statistics. Journal of Statistical Software, 31(10). Available
      from http://www.jstatsoft.org/v31/i10
    """

    return _var(alpha, True, axis=axis, w=w, d=d, angvar=angvar)


def _std(data, nanrobust, axis=None, deg=False, varroot=False):
    # convert to radians
    if deg:
        data = np.radians(data)

    # compute mean resultant vector length
    if nanrobust:
        R = nanresvec(data, axis=axis)
    else:
        R = resvec(data, axis=axis)

    # if just compute the square root, compute the variance and then
    # take the square root
    if varroot:
        V = 1 - R
        sd = np.sqrt(V)
    # otherwise calculate true circular standard deviation
    else:
        sd = np.sqrt(-2 * np.log(R))

    # convert back to degrees, if necessary
    if deg:
        sd = np.degrees(sd)

    return sd


def std(data, axis=None, deg=False, varroot=False, nancheck=True):
    """Computes circular standard deviation for circular data.
    Parameters
    ----------
    data : np.ndarray
        The array of angles
    axis : int (default=None)
        The axis along which to take the mean
    deg : bool (default=False)
        Whether the data is in degrees or radians
    varroot : bool (default=False)
        If True, compute the square root of the variance.  If False,
        compute the circlar standard deviation.
    nancheck : bool (default=True):
        Whether to check for NaNs in the input array
    Returns
    -------
    out : np.ndarray
        The calculated circular standard deviation
    References
    ----------
    http://en.wikipedia.org/wiki/Directional_statistics#Measures_of_location_and_spread
    """

    if nancheck and np.isnan(data).any():
        raise (ValueError,
               "NaN(s) detected in the input array.  If you want "
               "to ignore this message, please set nancheck to "
               "False.  If you want to make the stddev robust to "
               "NaNs, please use the nanstd function instead.")
    return _std(data, False, axis=axis, deg=deg, varroot=varroot)


def nanstd(data, axis=None, deg=False, varroot=False):
    """Computes circular standard deviation for circular data and is
    robust to invalid entries in the array.
    Parameters
    ----------
    data : np.ndarray
        The array of angles
    axis : int (default=None)
        The axis along which to take the mean
    deg : bool (default=False)
        Whether the data is in degrees or radians
    varroot : bool (default=False)
        If True, compute the square root of the variance.  If False,
        compute the circlar standard deviation.
    Returns
    -------
    out : np.ndarray
        The calculated circular standard deviation
    References
    ----------
    http://en.wikipedia.org/wiki/Directional_statistics#Measures_of_location_and_spread
    """

    return _std(data, True, axis=axis, deg=deg, varroot=varroot)


def _confmean(alpha, nanrobust, axis=None, ci=0.95, w=None, d=0):
    if w is None:
        w = np.ones(alpha.shape)

    # compute ingredients for conf. lim.
    if nanrobust:
        r = nanresvec(alpha, axis=axis, w=w, d=d)
    else:
        r = resvec(alpha, axis=axis, w=w, d=d)
    n = np.sum(w, axis=axis)
    R = n * r
    c2 = chi2.isf(1-ci, 1)

    # check for resultant vector length and select appropriate formula
    t = np.zeros(r.shape)

    tscalar = np.isscalar(r)
    if tscalar:
        r = np.array([r])
        n = np.array([n])
        R = np.array([R])
        t = np.array([t])

    # fill in values
    i = (r < 0.9) & (r > np.sqrt(c2 / 2. / n))
    t[i] = np.sqrt((2 * n[i] * (2 * R[i]**2 - n[i] * c2)) / (4 * n[i] - c2))

    j = r >= 0.9
    t[j] = np.sqrt(n[j] ** 2 - (n[j] ** 2 - R[j] ** 2) * np.exp(c2 / n[j]))

    t[~(i | j)] = np.nan

    # apply final transform
    t = np.arccos(t / R)
    if tscalar:
        t = t[0]

    return t


def confmean(alpha, axis=None, ci=0.95, w=None, d=0, nancheck=True):
    """Calculate the confidence limit of the mean of circular data.
    Parameters
    ----------
    alpha : np.ndarray
        The array of angles
    axis : int (default=None)
        The axis along which to take the mean
    ci : number (default=0.95)
        The confidence interval, between 0 and 1
    w : np.ndarray (default=None)
        An array of weights, if a weighted mean is desired
    d : number (default=0)
        The bin spacing, if the data is binned
    nancheck : bool (default=True)
        Whether or not to check for NaNs in the array
    Returns
    -------
    out : np.ndarray
        The confidence limit(s) of the input array
    References
    ----------
    Berens, P. (2009). CircStat: A MATLAB Toolbox for Circular
      Statistics. Journal of Statistical Software, 31(10). Available
      from http://www.jstatsoft.org/v31/i10
    """

    if nancheck and np.isnan(alpha).any():
        raise(ValueError,
              "NaN(s) detected in the input array.  If you want "
              "to ignore this message, please set nancheck to False.  "
              "If you want to make the mean robust to NaNs, please use "
              "the nanconfmean function instead.")

    return _confmean(alpha, False, axis=axis, ci=ci, w=w, d=d)


def nanconfmean(alpha, axis=None, ci=0.95, w=None, d=0):
    """Calculate the confidence limit of the mean of circular data.
    Parameters
    ----------
    alpha : np.ndarray
        The array of angles
    axis : int (default=None)
        The axis along which to take the mean
    ci : number (default=0.95)
        The confidence interval, between 0 and 1
    w : np.ndarray (default=None)
        An array of weights, if a weighted mean is desired
    d : number (default=0)
        The bin spacing, if the data is binned
    Returns
    -------
    out : np.ndarray
        The confidence limit(s) of the input array
    References
    ----------
    Berens, P. (2009). CircStat: A MATLAB Toolbox for Circular
      Statistics. Journal of Statistical Software, 31(10). Available
      from http://www.jstatsoft.org/v31/i10
    """

    return _confmean(alpha, True, axis=axis, ci=ci, w=w, d=d)


def _corrcc(alpha1, alpha2, nanrobust, axis=None):
    if axis is not None and alpha1.shape[axis] != alpha2.shape[axis]:
        raise(ValueError, "shape mismatch")

    # compute mean directions
    if axis is None:
        n = alpha1.size
    else:
        n = alpha1.shape[axis]

    #################################################################
    c1 = np.cos(alpha1)
    c1_2 = np.cos(2*alpha1)
    c2 = np.cos(alpha2)
    c2_2 = np.cos(2*alpha2)
    s1 = np.sin(alpha1)
    s1_2 = np.sin(2*alpha1)
    s2 = np.sin(alpha2)
    s2_2 = np.sin(2*alpha2)

    if nanrobust:
        sumfunc = lambda x: np.nansum(x, axis=axis)
    else:
        sumfunc = lambda x: np.sum(x, axis=axis)

    num = 4 * (sumfunc(c1*c2) * sumfunc(s1*s2) -
               sumfunc(c1*s2) * sumfunc(s1*c2))
    den = np.sqrt((n**2 - sumfunc(c1_2)**2 - sumfunc(s1_2)**2) *
                  (n**2 - sumfunc(c2_2)**2 - sumfunc(s2_2)**2))

    rho = num / den

    return rho


def corrcc(alpha1, alpha2, axis=None, nancheck=True):
    """Calculate the correlation coefficient between two sets of
    circular data.
    Parameters
    ----------
    alpha1 : np.ndarray
        The first set of circular data
    alpha2 : np.ndarray
        The second set of circular data
    axis : int (default=None)
        The axis along which to take the mean
    nancheck : bool (default=True)
        Whether or not to check for NaNs in the array
    Returns
    -------
    out : tuple (np.ndarray, np.ndarray)
        The correlation coefficient(s) and the p-value(s)
    References
    ----------
    Berens, P. (2009). CircStat: A MATLAB Toolbox for Circular
      Statistics. Journal of Statistical Software, 31(10). Available
      from http://www.jstatsoft.org/v31/i10
    Fisher, N. & Lee, A. (1983). A correlation coefficient for circular
      data. Biometrika. 70 (2), 327-332.
    """

    if nancheck and (np.isnan(alpha1).any() or np.isnan(alpha2).any()):
        raise ValueError("NaN(s) detected in the input array. "
                         "If you want to ignore this message, please set "
                         "nancheck to False. If you want to make the mean "
                         "robust to NaNs, please use the nancorrcc function "
                         "instead.")

    return _corrcc(alpha1, alpha2, False, axis=axis)


def nancorrcc(alpha1, alpha2, axis=None):
    """Calculate the correlation coefficient between two sets of
    circular data.
    Parameters
    ----------
    alpha1 : np.ndarray
        The first set of circular data
    alpha2 : np.ndarray
        The second set of circular data
    axis : int (default=None)
        The axis along which to take the mean
    Returns
    -------
    out : tuple (np.ndarray, np.ndarray)
        The correlation coefficient(s) and the p-value(s)
    References
    ----------
    Berens, P. (2009). CircStat: A MATLAB Toolbox for Circular
      Statistics. Journal of Statistical Software, 31(10). Available
      from http://www.jstatsoft.org/v31/i10
    Fisher, N. & Lee, A. (1983). A correlation coefficient for circular
      data. Biometrika. 70 (2), 327-332.
    """

    return _corrcc(alpha1, alpha2, True, axis=axis)


def _vmpar(alpha, nanrobust, axis=None, w=None, d=0):
    if w is None:
        w = np.ones(alpha.shape)

    if nanrobust:
        _kappa = nankappa(alpha, axis=axis, w=w, d=d)
        _thetahat = nanmean(alpha, axis=axis, w=w)
    else:
        _kappa = kappa(alpha, axis=axis, w=w, d=d)
        _thetahat = mean(alpha, axis=axis, w=w)

    return (_thetahat, _kappa)


def vmpar(alpha, axis=None, w=None, d=0, nancheck=True):
    """Estimate the parameters of a Von Mises distribution.
    Parameters
    ----------
    alpha : np.ndarray
        The array of angles
    axis : int (default=None)
        The axis along which to take the mean
    w : np.ndarray (default=None)
        An array of weights, if a weighted mean is desired
    d : number (default=0)
        The bin spacing, if the data is binned
    nancheck : bool (default=True)
        Whether or not to check for NaNs in the array
    Returns
    -------
    out : tuple (np.ndarray, np.ndarray)
        The theta and kappa parameters, respectively
    References
    ----------
    Berens, P. (2009). CircStat: A MATLAB Toolbox for Circular
      Statistics. Journal of Statistical Software, 31(10). Available
      from http://www.jstatsoft.org/v31/i10
    """

    if nancheck and np.isnan(alpha).any():
        raise(ValueError,
              "NaN(s) detected in the input array.  "
              "If you want to ignore this message, please "
              "set nancheck to False.  If you want to make "
              "the mean robust to NaNs, please use the nanvmpar "
              "function instead.")

    return _vmpar(alpha, False, axis=axis, w=w, d=d)


def nanvmpar(alpha, axis=None, w=None, d=0):
    """Estimate the parameters of a Von Mises distribution.
    Parameters
    ----------
    alpha : np.ndarray
        The array of angles
    axis : int (default=None)
        The axis along which to take the mean
    w : np.ndarray (default=None)
        An array of weights, if a weighted mean is desired
    d : number (default=0)
        The bin spacing, if the data is binned
    Returns
    -------
    out : tuple (np.ndarray, np.ndarray)
        The theta and kappa parameters, respectively
    References
    ----------
    Berens, P. (2009). CircStat: A MATLAB Toolbox for Circular
      Statistics. Journal of Statistical Software, 31(10). Available
      from http://www.jstatsoft.org/v31/i10
    """

    return _vmpar(alpha, True, axis=axis, w=w, d=d)


def vmlogpdf(alpha, thetahat=0.0, kappa=1.0):
    """Computes the circular Von Mises log-PDF with preferred
    direction thetahat and concentration kappa at each of the angles
    in alpha.
    The vmpdf is given by f(phi) =
    (1/(2pi*I0(kappa))*exp(kappa*cos(phi-thetahat)
    Parameters
    ----------
    alpha : np.ndarray
        The array of angles
    thetahat : number
        The preferred direction
    kappa : number
        Concentration at each of the angles in alpha
    Returns
    -------
    out : np.ndarray
        The Von Mises log-PDF for alpha
    References
    ----------
    Berens, P. (2009). CircStat: A MATLAB Toolbox for Circular
      Statistics. Journal of Statistical Software, 31(10). Available
      from http://www.jstatsoft.org/v31/i10
    """

    C = -np.log(2 * np.pi * iv(0, kappa))
    p = C + (kappa * np.cos(alpha - thetahat))
    return p


def vmpdf(alpha, thetahat=0.0, kappa=1.0):
    """Computes the circular Von Mises PDF with preferred direction
    thetahat and concentration kappa at each of the angles in alpha.
    The vmpdf is given by f(phi) =
    (1/(2pi*I0(kappa))*exp(kappa*cos(phi-thetahat)
    Parameters
    ----------
    alpha : np.ndarray
        The array of angles
    thetahat : number
        The preferred direction
    kappa : number
        Concentration at each of the angles in alpha
    Returns
    -------
    out : np.ndarray
        The Von Mises PDF for alpha
    References
    ----------
    Berens, P. (2009). CircStat: A MATLAB Toolbox for Circular
      Statistics. Journal of Statistical Software, 31(10). Available
      from http://www.jstatsoft.org/v31/i10
    """

    p = np.exp(vmlogpdf(alpha, thetahat=thetahat, kappa=kappa))
    return p


def _kappa(alpha, nanrobust, axis=None, w=None, d=0):
    if w is None:
        w = np.ones(alpha.shape)

    if axis is None:
        n = alpha.size
    else:
        n = alpha.shape[axis]

    if nanrobust:
        R = nanresvec(alpha, axis=axis, w=w, d=d)
    else:
        R = resvec(alpha, axis=axis, w=w, d=d)

    if type(R) is not np.ndarray or len(R.shape) == 0:
        R = np.array([R])

    i1 = R < 0.53
    i2 = (R >= 0.53) & (R < 0.85)
    i3 = ~(i1 | i2)

    kappa = np.empty(R.shape)
    kappa[i1] = 2*R[i1] + R[i1]**3 + 5*R[i1]**5/6
    kappa[i2] = -.4 + 1.39*R[i2] + 0.43/(1-R[i2])
    kappa[i3] = 1/(R[i3]**3 - 4*R[i3]**2 + 3*R[i3])

    if (n < 15) & (n > 1):
        i1 = kappa < 2
        i2 = ~i1
        kappa[i1] = np.max(np.concatenate((kappa[i1]-2*(n*kappa[i1])**-1,
                                           np.zeros(i1.shape)),
                                          axis=i1.ndim-1),
                           axis=i1.ndim-1)
        kappa[i2] = (n-1)**3 * kappa[i2] / (n**3 + n)

    return kappa


def kappa(alpha, axis=None, w=None, d=0, nancheck=True):
    """Computes an approximation to the ML estimate of the
    concentration parameter kappa of the von Mises distribution.
    Parameters
    ----------
    alpha : np.ndarray
        The array of angles
    axis : int (default=None)
        The axis along which to take the mean
    w : np.ndarray (default=None)
        An array of weights, if a weighted mean is desired
    nancheck : bool (default=True)
        Whether or not to check for NaNs in the array
    Returns
    -------
    out : np.ndarray
        Kappa parameter estimations
    References
    ----------
    Berens, P. (2009). CircStat: A MATLAB Toolbox for Circular
      Statistics. Journal of Statistical Software, 31(10). Available
      from http://www.jstatsoft.org/v31/i10
    """

    if nancheck and np.isnan(alpha).any():
        raise(ValueError,
              "NaN(s) detected in the input array.  If "
              "you want to ignore this message, please set "
              "nancheck to False.  If you want to make the "
              "mean robust to NaNs, please use the nankappa "
              "function instead.")

    return _kappa(alpha, False, axis=axis, w=w, d=d)


def nankappa(alpha, axis=None, w=None, d=0):
    """Computes an approximation to the ML estimate of the
    concentration parameter kappa of the von Mises distribution.
    Parameters
    ----------
    alpha : np.ndarray
        The array of angles
    axis : int (default=None)
        The axis along which to take the mean
    w : np.ndarray (default=None)
        An array of weights, if a weighted mean is desired
    Returns
    -------
    out : np.ndarray
        Kappa parameter estimations
    References
    ----------
    Berens, P. (2009). CircStat: A MATLAB Toolbox for Circular
      Statistics. Journal of Statistical Software, 31(10). Available
      from http://www.jstatsoft.org/v31/i10
    """

    return _kappa(alpha, True, axis=axis, w=w, d=d)




try:
    from .matplotlibmisc import *
except ImportError as e:

    try:
        from .matplotlibmisc import wes_palettes, rgb_to_hex, hex_to_rgb
        print('succeeded to install wes_palettes, rgb_to_hex, hex_to_rgb from script')
        print('failed to install discrete_colormap, create_colormap, shiftedColorMap from script')

    except ImportError as e:
        warn = 'failed to install discrete_colormap, create_colormap, shiftedColorMap from script, wes_palettes, {}'
        print(warn.format('rgb_to_hex, hex_to_rgb from script'))

from .misc import *
from .inherit_docstrings import InheritableDocstrings

